import depthai as dai
import numpy as np

from ..messages.creators import create_segmentation_message


class SegmentationParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the segmentation models.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    background_class : bool
        Whether to add additional layer for background.

    Output Message/s
    ----------------
    **Type**: dai.ImgFrame

    **Description**: Segmentation message containing the segmentation mask. Every pixel belongs to exactly one class.

    Error Handling
    --------------
    **ValueError**: If the number of output layers is not E{1}.

    **ValueError**: If the number of dimensions of the output tensor is not E{3}.
    """

    def __init__(self, background_class=False):
        """Initializes the SegmentationParser node.

        @param background_class: Whether to add additional layer for background.
        @type background_class: bool
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()
        self.background_class = background_class

    def setBackgroundClass(self, background_class):
        """Sets the background class.

        @param background_class: Whether to add additional layer for background.
        @type background_class: bool
        """
        self.background_class = background_class

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()

            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )

            segmentation_mask = output.getTensor(output_layer_names[0], dequantize=True)
            if len(segmentation_mask.shape) == 4:
                segmentation_mask = segmentation_mask[0]
            else:
                segmentation_mask = segmentation_mask.transpose(2, 0, 1)

            if len(segmentation_mask.shape) != 3:
                raise ValueError(
                    f"Expected 3D output tensor, got {len(segmentation_mask.shape)}D."
                )

            if self.background_class:
                segmentation_mask = np.vstack(
                    (
                        np.zeros(
                            (1, segmentation_mask.shape[1], segmentation_mask.shape[2]),
                            dtype=np.float32,
                        ),
                        segmentation_mask,
                    )
                )

            class_map = (
                np.argmax(segmentation_mask, axis=0)
                .reshape(segmentation_mask.shape[1], segmentation_mask.shape[2], 1)
                .astype(np.uint8)
            )

            imgFrame = create_segmentation_message(class_map)
            imgFrame.setTimestamp(output.getTimestamp())
            self.out.send(imgFrame)



import depthai as dai

from .mapping import parser_mapping


def visualize(
    frame: dai.ImgFrame, message: dai.Buffer, parser_name: str, extraParams: dict
):
    """Calls the appropriate visualizer based on the parser name and returns True if the
    pipeline should be stopped."""
    visualizer = parser_mapping[parser_name]
    return visualizer(frame, message, extraParams)



from .classification import create_classification_message
from .depth import create_depth_message
from .detection import create_detection_message, create_line_detection_message
from .image import create_image_message
from .keypoints import create_hand_keypoints_message, create_keypoints_message
from .misc import create_age_gender_message
from .segmentation import create_sam_message, create_segmentation_message
from .thermal import create_thermal_message
from .tracked_features import create_tracked_features_message

__all__ = [
    "create_image_message",
    "create_segmentation_message",
    "create_hand_keypoints_message",
    "create_detection_message",
    "create_depth_message",
    "create_line_detection_message",
    "create_tracked_features_message",
    "create_keypoints_message",
    "create_thermal_message",
    "create_classification_message",
    "create_sam_message",
    "create_age_gender_message",
]



import cv2
import depthai as dai
import numpy as np

from .colors import get_adas_colors, get_ewasr_colors, get_selfie_colors
from .messages import parse_segmentation_message


def visualize_segmentation(
    frame: dai.ImgFrame, message: dai.ImgFrame, extraParams: dict
):
    mask = parse_segmentation_message(message)
    frame = cv2.resize(frame, (mask.shape[1], mask.shape[0]))

    n_classes = extraParams.get("n_classes", None)

    if n_classes is None:
        raise ValueError("Number of classes not provided in NN archive metadata.")

    if n_classes == 2:
        COLORS = get_selfie_colors()
    elif n_classes == 3:
        COLORS = get_ewasr_colors()
    else:
        COLORS = get_adas_colors()

    colormap = np.zeros((mask.shape[0], mask.shape[1], 3), dtype=np.uint8)
    for class_id, color in enumerate(COLORS):
        m = mask == class_id
        colormap[m] = color

    alpha = 0.5
    overlay = cv2.addWeighted(colormap, alpha, frame, 1 - alpha, 0)

    cv2.imshow("Segmentation", overlay)

    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import depthai as dai

from depthai_nodes.ml.parsers import (
    ClassificationParser,
    KeypointParser,
    MonocularDepthParser,
    MPPalmDetectionParser,
    SCRFDParser,
    SegmentationParser,
    XFeatParser,
    YOLOExtendedParser,
)


def setup_scrfd_parser(parser: SCRFDParser, params: dict):
    """Setup the SCRFD parser with the required metadata."""
    try:
        num_anchors = params["num_anchors"]
        feat_stride_fpn = params["feat_stride_fpn"]
        parser.setNumAnchors(num_anchors)
        parser.setFeatStrideFPN(feat_stride_fpn)
    except Exception:
        print(
            "This NN archive does not have required metadata for SCRFDParser. Skipping setup..."
        )


def setup_segmentation_parser(parser: SegmentationParser, params: dict):
    """Setup the segmentation parser with the required metadata."""
    try:
        background_class = params["background_class"]
        parser.setBackgroundClass(background_class)
    except Exception:
        print(
            "This NN archive does not have required metadata for SegmentationParser. Skipping setup..."
        )


def setup_keypoint_parser(parser: KeypointParser, params: dict):
    """Setup the keypoint parser with the required metadata."""
    try:
        num_keypoints = params["n_keypoints"]
        scale_factor = params["scale_factor"]
        parser.setNumKeypoints(num_keypoints)
        parser.setScaleFactor(scale_factor)
    except Exception:
        print(
            "This NN archive does not have required metadata for KeypointParser. Skipping setup..."
        )


def setup_classification_parser(parser: ClassificationParser, params: dict):
    """Setup the classification parser with the required metadata."""
    try:
        classes = params["classes"]
        is_softmax = params["is_softmax"]
        parser.setClasses(classes)
        parser.setSoftmax(is_softmax)
    except Exception:
        print(
            "This NN archive does not have required metadata for ClassificationParser. Skipping setup..."
        )


def setup_monocular_depth_parser(parser: MonocularDepthParser, params: dict):
    """Setup the monocular depth parser with the required metadata."""
    try:
        depth_type = params["depth_type"]
        depth_limit = params["depth_limit"]
        if depth_type == "relative":
            parser.setRelativeDepthType()
        else:
            parser.setMetricDepthType()
        parser.setDepthLimit(depth_limit)
    except Exception:
        print(
            "This NN archive does not have required metadata for MonocularDepthParser. Skipping setup..."
        )


def setup_xfeat_parser(parser: XFeatParser, params: dict):
    """Setup the XFeat parser with the required metadata."""
    try:
        input_size = params["input_size"]
        parser.setInputSize(input_size)
        parser.setOriginalSize(input_size)
    except Exception:
        print(
            "This NN archive does not have required metadata for XFeatParser. Skipping setup..."
        )


def setup_yolo_extended_parser(parser: YOLOExtendedParser, params: dict):
    """Setup the YOLO parser with the required metadata."""
    try:
        n_classes = params["n_classes"]
        parser.setNumClasses(n_classes)
    except Exception:
        print(
            "This NN archive does not have required metadata for YOLOExtendedParser. Skipping setup..."
        )


def setup_palm_detection_parser(parser: MPPalmDetectionParser, params: dict):
    """Setup the Palm Detection parser with the required metadata."""
    try:
        scale = params["scale"]
        parser.setScale(scale)
    except Exception:
        print(
            "This NN archive does not have required metadata for MPPalmDetectionParser. Skipping setup..."
        )


def setup_parser(parser: dai.ThreadedNode, nn_archive: dai.NNArchive, parser_name: str):
    """Setup the parser with the NN archive."""

    extraParams = (
        nn_archive.getConfig().getConfigV1().model.heads[0].metadata.extraParams
    )

    if parser_name == "SCRFDParser":
        setup_scrfd_parser(parser, extraParams)
    elif parser_name == "SegmentationParser":
        setup_segmentation_parser(parser, extraParams)
    elif parser_name == "KeypointParser":
        setup_keypoint_parser(parser, extraParams)
    elif parser_name == "ClassificationParser":
        setup_classification_parser(parser, extraParams)
    elif parser_name == "MonocularDepthParser":
        setup_monocular_depth_parser(parser, extraParams)
    elif parser_name == "XFeatParser":
        setup_xfeat_parser(parser, extraParams)
    elif parser_name == "YOLOExtendedParser":
        setup_yolo_extended_parser(parser, extraParams)
    elif parser_name == "MPPalmDetectionParser":
        setup_palm_detection_parser(parser, extraParams)



from typing import List

import depthai as dai
import numpy as np

from ..messages.creators import create_classification_message


class ClassificationParser(dai.node.ThreadedHostNode):
    """Postprocessing logic for Classification model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    classes : List[str]
        List of class names to be used for linking with their respective scores. Expected to be in the same order as Neural Network's output. If not provided, the message will only return sorted scores.
    is_softmax : bool = True
        If False, the scores are converted to probabilities using softmax function.
    n_classes : int = len(classes)
        Number of provided classes. This variable is set automatically based on provided classes.

    Output Message/s
    ----------------
    **Type** : Classifications(dai.Buffer):
         An object with attributes `classes` and `scores`. `classes` is a list of classes, sorted in descending order of scores. `scores` is a list of corresponding scores.
    """

    def __init__(self, classes: List[str] = None, is_softmax: bool = True):
        """Initializes the ClassificationParser node.

        @param classes: List of class names to be used for linking with their respective
            scores.
        @param is_softmax: If False, the scores are converted to probabilities using
            softmax function.
        """

        dai.node.ThreadedHostNode.__init__(self)
        self.out = self.createOutput()
        self.input = self.createInput()
        self.classes = classes if classes is not None else []
        self.n_classes = len(self.classes)
        self.is_softmax = is_softmax

    def setClasses(self, classes: List[str]):
        """Sets the class names for the classification model.

        @param classes: List of class names to be used for linking with their respective
            scores.
        """
        self.classes = classes if classes is not None else []
        self.n_classes = len(self.classes)

    def setSoftmax(self, is_softmax: bool):
        """Sets the softmax flag for the classification model.

        @param is_softmax: If False, the parser will convert the scores to probabilities
            using softmax function.
        """
        self.is_softmax = is_softmax

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()
            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )

            if self.n_classes == 0:
                raise ValueError("Classes must be provided for classification.")

            scores = output.getTensor(output_layer_names[0], dequantize=True).astype(
                np.float32
            )
            scores = np.array(scores).flatten()

            if len(scores) != self.n_classes and self.n_classes != 0:
                raise ValueError(
                    f"Number of labels and scores mismatch. Provided {self.n_classes} class names and {len(scores)} scores."
                )

            if not self.is_softmax:
                ex = np.exp(scores)
                scores = ex / np.sum(ex)

            msg = create_classification_message(self.classes, scores)
            msg.setTimestamp(output.getTimestamp())

            self.out.send(msg)



from typing import List, Union

import numpy as np

from ...messages import Classifications


def create_classification_message(
    classes: List, scores: Union[np.ndarray, List]
) -> Classifications:
    """Create a message for classification. The message contains the class names and
    their respective scores, sorted in descending order of scores.

    @param classes: A list containing class names.
    @type classes: List
    @param scores: A numpy array of shape (n_classes,) containing the probability score of each class.
    @type scores: np.ndarray

    @return: A message with attributes `classes` and `scores`. `classes` is a list of classes, sorted in descending order of scores. `scores` is a list of the corresponding scores.
    @rtype: Classifications

    @raises ValueError: If the provided classes are None.
    @raises ValueError: If the provided classes are not a list.
    @raises ValueError: If the provided classes are empty.
    @raises ValueError: If the provided scores are None.
    @raises ValueError: If the provided scores are not a list or a numpy array.
    @raises ValueError: If the provided scores are empty.
    @raises ValueError: If the provided scores are not a 1D array.
    @raises ValueError: If the provided scores are not of type float.
    @raises ValueError: If the provided scores do not sum to 1.
    @raises ValueError: If the number of labels and scores mismatch.
    """
    if isinstance(classes, type(None)):
        raise ValueError("Classes should not be None.")

    if not isinstance(classes, list):
        raise ValueError(f"Classes should be a list, got {type(classes)}.")

    if len(classes) == 0:
        raise ValueError("Classes should not be empty.")

    if type(scores) == type(None):
        raise ValueError("Scores should not be None.")

    if not isinstance(scores, np.ndarray) and not isinstance(scores, list):
        raise ValueError(
            f"Scores should be a list or a numpy array, got {type(scores)}."
        )

    if isinstance(scores, list):
        scores = np.array(scores)

    if len(scores) == 0:
        raise ValueError("Scores should not be empty.")

    if len(scores) != len(scores.flatten()):
        raise ValueError(f"Scores should be a 1D array, got {scores.shape}.")

    scores = scores.flatten()

    if not np.issubdtype(scores.dtype, np.floating):
        raise ValueError(f"Scores should be of type float, got {scores.dtype}.")

    if not np.isclose(np.sum(scores), 1.0, atol=1e-2):
        raise ValueError(f"Scores should sum to 1, got {np.sum(scores)}.")

    if len(scores) != len(classes):
        raise ValueError(
            f"Number of labels and scores mismatch. Provided {len(scores)} scores and {len(classes)} class names."
        )

    classification_msg = Classifications()
    sorted_args = np.argsort(-scores, kind="stable")
    scores = scores[sorted_args]

    classification_msg.classes = [classes[i] for i in sorted_args]
    classification_msg.scores = scores.tolist()

    return classification_msg



from typing import List, Tuple, Union

import depthai as dai
import numpy as np


class ImgDetectionExtended(dai.ImgDetection):
    """ImgDetectionExtended class for storing image detection with keypoints and masks.

    Attributes
    ----------
    keypoints: Union[List[Tuple[float, float]], List[Tuple[float, float, float]]]
        Keypoints of the image detection.
    masks: np.ndarray
        Mask of the image segmentation.
    """

    def __init__(self):
        """Initializes the ImgDetectionExtended object."""
        dai.ImgDetection.__init__(self)  # TODO: change to super().__init__()?
        self._keypoints: Union[
            List[Tuple[float, float]], List[Tuple[float, float, float]]
        ] = []
        self._mask: np.ndarray = np.array([])

    @property
    def keypoints(
        self,
    ) -> Union[List[Tuple[float, float]], List[Tuple[float, float, float]]]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: Union[List[Tuple[float, float]], List[Tuple[float, float, float]]]
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(
        self,
        value: Union[
            List[Tuple[Union[int, float], Union[int, float]]],
            List[Tuple[Union[int, float], Union[int, float], Union[int, float]]],
        ],
    ):
        """Sets the keypoints.

        @param value: List of keypoints.
        @type value: Union[List[Tuple[Union[int, float], Union[int, float]]],
            List[Tuple[Union[int, float], Union[int, float], Union[int, float]]]]
        @raise TypeError: If the keypoints are not a list.
        @raise TypeError: If each keypoint is not a tuple of two or three floats or
            integers.
        """
        if not isinstance(value, list):
            raise TypeError("Keypoints must be a list")
        dim = len(value[0]) if value else 0
        for item in value:
            if (
                not (isinstance(item, tuple) or isinstance(item, list))
                or (len(item) != 2 and len(item) != 3)
                or not all(isinstance(i, (int, float)) for i in item)
            ):
                raise TypeError(
                    "Each keypoint must be a tuple of two or three floats or integers."
                )
            if len(item) != dim:
                raise ValueError(
                    "All keypoints must be of the same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
                )
        keypoints = []
        for items in value:
            keypoints.append(tuple(float(v) for v in items))
        self._keypoints = keypoints

    @property
    def mask(self) -> np.ndarray:
        """Returns the mask coefficients.

        @return: Mask coefficients.
        @rtype: np.ndarray
        """
        return self._mask

    @mask.setter
    def mask(self, value: np.ndarray):
        """Sets the mask coefficients.

        @param value: Mask coefficients.
        @type value: np.ndarray
        @raise TypeError: If the mask is not a numpy array.
        @raise ValueError: If the mask is not of shape (H/4, W/4).
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Mask must be a numpy array")
        if len(value.shape) != 2:
            raise ValueError("Mask must be of shape (H/4, W/4)")
        self._mask = value


class ImgDetectionsExtended(dai.Buffer):
    """ImgDetectionsExtended class for storing image detections with keypoints.

    Attributes
    ----------
    detections: List[ImgDetectionExtended]
        Image detections with keypoints.
    """

    def __init__(self):
        """Initializes the ImgDetectionsExtended object."""
        dai.Buffer.__init__(self)  # TODO: change to super().__init__()?
        self._detections: List[ImgDetectionExtended] = []

    @property
    def detections(self) -> List[ImgDetectionExtended]:
        """Returns the image detections with keypoints.

        @return: List of image detections with keypoints.
        @rtype: List[ImgDetectionExtended]
        """
        return self._detections

    @detections.setter
    def detections(self, value: List[ImgDetectionExtended]):
        """Sets the image detections with keypoints.

        @param value: List of image detections with keypoints.
        @type value: List[ImgDetectionExtended]
        @raise TypeError: If the detections are not a list.
        @raise TypeError: If each detection is not an instance of ImgDetectionExtended.
        """
        if not isinstance(value, list):
            raise TypeError("Detections must be a list")
        for item in value:
            if not isinstance(item, ImgDetectionExtended):
                raise TypeError(
                    "Each detection must be an instance of ImgDetectionExtended"
                )
        self._detections = value



from typing import List

from ...messages import AgeGender, Classifications


def create_age_gender_message(age: float, gender_prob: List[float]) -> AgeGender:
    """Create a DepthAI message for the age and gender probability.

    @param age: Detected person age (must be multiplied by 100 to get years).
    @type age: float
    @param gender_prob: Detected person gender probability [female, male].
    @type gender_prob: List[float]
    @return: AgeGender message containing the predicted person's age and Classifications
        message containing the classes and probabilities of the predicted gender.
    @rtype: AgeGender
    @raise ValueError: If age is not a float.
    @raise ValueError: If gender_prob is not a list.
    @raise ValueError: If each item in gender_prob is not a float.
    """

    if not isinstance(age, float):
        raise ValueError(f"Age should be float, got {type(age)}.")

    if not isinstance(gender_prob, List):
        raise ValueError(f"Gender_prob should be list, got {type(gender_prob)}.")
    if len(gender_prob) != 2:
        raise ValueError(
            f"Gender_prob list should have 2 values, got {len(gender_prob)}."
        )

    for item in gender_prob:
        if not isinstance(item, (float)):
            raise ValueError(
                f"Gender_prob list values must be of type float, instead got {type(item)}."
            )

    if sum(gender_prob) < 0.99 or sum(gender_prob) > 1.01:
        raise ValueError(
            f"Gender_prob list must contain probabilities and sum to 1, got sum {sum(gender_prob)}."
        )

    age_gender_message = AgeGender()
    age_gender_message.age = age
    gender = Classifications()
    gender.classes = ["female", "male"]
    gender.scores = gender_prob
    age_gender_message.gender = gender

    return age_gender_message



from typing import List, Tuple, Union

import depthai as dai
import numpy as np

from ...messages import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
    Line,
    Lines,
)


def create_detection_message(
    bboxes: np.ndarray,
    scores: np.ndarray,
    labels: List[int] = None,
    keypoints: Union[
        List[Tuple[float, float]], List[Tuple[float, float, float]]
    ] = None,
    masks: List[np.ndarray] = None,
) -> dai.ImgDetections:
    """Create a DepthAI message for an object detection.

    @param bbox: Bounding boxes of detected objects of shape (N,4) meaning [...,[x_min, y_min, x_max, y_max],...].
    @type bbox: np.ndarray
    @param scores: Confidence scores of detected objects of shape (N,).
    @type scores: np.ndarray
    @param labels: Labels of detected objects of shape (N,).
    @type labels: List[int]
    @param keypoints: Keypoints of detected objects of shape (N,2) or (N,3).
    @type keypoints: Optional[Union[List[Tuple[float, float]], List[Tuple[float, float, float]]]]
    @param masks: Masks of detected objects of shape (N, H, W).
    @type masks: List[np.ndarray]

    @return: Message containing the bounding boxes, labels, confidence scores, and keypoints of detected objects.
    @rtype: dai.ImgDetections OR ImgDetectionsExtended

    @raise ValueError: If the bboxes are not a numpy array.
    @raise ValueError: If the bboxes are not of shape (N,4).
    @raise ValueError: If the bboxes 2nd dimension is not of size 4.
    @raise ValueError: If the bboxes are not in format [x_min, y_min, x_max, y_max] where xmin < xmax and ymin < ymax.
    @raise ValueError: If the scores are not a numpy array.
    @raise ValueError: If the scores are not of shape (N,).
    @raise ValueError: If the scores do not have the same length as bboxes.
    @raise ValueError: If the labels are not a list.
    @raise ValueError: If each label is not an integer.
    @raise ValueError: If the labels do not have the same length as bboxes.
    @raise ValueError: If the keypoints are not a list.
    @raise ValueError: If each keypoint pair is not a tuple of two or three floats.
    @raise ValueError: If the keypoints do not have the same length as bboxes.
    @raise ValueError: If the masks are not a list.
    @raise ValueError: If each mask is not a 2D numpy array.
    """

    # checks for bboxes
    if not isinstance(bboxes, np.ndarray):
        raise ValueError(f"Bounding boxes should be numpy array, got {type(bboxes)}.")

    if len(bboxes) != 0:
        if len(bboxes.shape) != 2:
            raise ValueError(
                f"Bounding boxes should be of shape (N,4) meaning [...,[x_min, y_min, x_max, y_max],...], got {bboxes.shape}."
            )
        if bboxes.shape[1] != 4:
            raise ValueError(
                f"Bounding boxes 2nd dimension should be of size 4 e.g. [x_min, y_min, x_max, y_max] got {bboxes.shape[1]}."
            )

        x_valid = bboxes[:, 0] < bboxes[:, 2]
        y_valid = bboxes[:, 1] < bboxes[:, 3]
        if not (np.all(x_valid) and np.all(y_valid)):
            raise ValueError(
                "Bounding boxes should be in format [x_min, y_min, x_max, y_max] where xmin < xmax and ymin < ymax."
            )

    # checks for scores
    if not isinstance(scores, np.ndarray):
        raise ValueError(f"Scores should be numpy array, got {type(scores)}.")

    if len(scores) != 0:
        if len(scores.shape) != 1:
            raise ValueError(
                f"Scores should be of shape (N,) meaning, got {scores.shape}."
            )
        if scores.shape[0] != bboxes.shape[0]:
            raise ValueError(
                f"Scores should have same length as bboxes, got {scores.shape[0]} and {bboxes.shape[0]}."
            )

    # checks for labels
    if labels is not None:
        if not isinstance(labels, List):
            raise ValueError(f"Labels should be list, got {type(labels)}.")
        for label in labels:
            if not isinstance(label, int):
                raise ValueError(
                    f"Labels should be list of integers, got {type(label)}."
                )
        if len(labels) != bboxes.shape[0]:
            raise ValueError(
                f"Labels should have same length as bboxes, got {len(labels)} and {bboxes.shape[0]}."
            )

    # checks for keypoints
    if keypoints is not None:
        if not isinstance(keypoints, List):
            raise ValueError(f"Keypoints should be list, got {type(keypoints)}.")
        if len(keypoints) != bboxes.shape[0]:
            raise ValueError(
                f"Keypoints should have same length as bboxes, got {len(keypoints)} and {bboxes.shape[0]}."
            )

        for object_keypoints in keypoints:
            dim = len(object_keypoints[0]) if len(object_keypoints) != 0 else 0
            for point in object_keypoints:
                if not isinstance(point, (Tuple, List)):
                    raise ValueError(
                        f"Keypoint pairs should be list of tuples, got {type(point)}."
                    )
                if len(point) not in [2, 3]:
                    raise ValueError(
                        f"Keypoint pairs should be list of tuples of length 2 or 3, got {len(point)}."
                    )
                if len(point) != dim:
                    raise ValueError(
                        "All keypoints should be of same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
                    )

    if masks is not None and len(masks) != 0:
        if not isinstance(masks, List):
            raise ValueError(f"masks should be list, got {type(masks)}.")
        for mask in masks:
            if not isinstance(mask, np.ndarray):
                raise ValueError(f"mask should be numpy array, got {type(mask)}.")
            if len(mask.shape) != 2:
                raise ValueError(f"mask should be a 2D matrix, got {mask.shape}.")

        if len(masks) != bboxes.shape[0]:
            raise ValueError(
                f"masks should have same length as bboxes, got {len(masks)} and {bboxes.shape[0]}."
            )

    if keypoints is not None or masks is not None:
        img_detection = ImgDetectionExtended
        img_detections = ImgDetectionsExtended
    else:
        img_detection = dai.ImgDetection
        img_detections = dai.ImgDetections

    detections = []
    for i in range(bboxes.shape[0]):
        detection = img_detection()
        detection.xmin = bboxes[i][0]
        detection.ymin = bboxes[i][1]
        detection.xmax = bboxes[i][2]
        detection.ymax = bboxes[i][3]
        detection.confidence = scores[i]
        if labels is None:
            detection.label = 0
        else:
            detection.label = labels[i]
        if keypoints is not None:
            detection.keypoints = keypoints[i]
        if masks is not None:
            detection.mask = masks[i]
        detections.append(detection)

    detections_msg = img_detections()
    detections_msg.detections = detections
    return detections_msg


def create_line_detection_message(lines: np.ndarray, scores: np.ndarray):
    """Create a DepthAI message for a line detection.

    @param lines: Detected lines of shape (N,4) meaning [...,[x_start, y_start, x_end, y_end],...].
    @type lines: np.ndarray
    @param scores: Confidence scores of detected lines of shape (N,).
    @type scores: np.ndarray

    @return: Message containing the lines and confidence scores of detected lines.
    @rtype: Lines

    @raise ValueError: If the lines are not a numpy array.
    @raise ValueError: If the lines are not of shape (N,4).
    @raise ValueError: If the lines 2nd dimension is not of size E{4}.
    @raise ValueError: If the scores are not a numpy array.
    @raise ValueError: If the scores are not of shape (N,).
    @raise ValueError: If the scores do not have the same length as lines.
    """

    # checks for lines
    if not isinstance(lines, np.ndarray):
        raise ValueError(f"Lines should be numpy array, got {type(lines)}.")
    if len(lines) != 0:
        if len(lines.shape) != 2:
            raise ValueError(
                f"Lines should be of shape (N,4) meaning [...,[x_start, y_start, x_end, y_end],...], got {lines.shape}."
            )
        if lines.shape[1] != 4:
            raise ValueError(
                f"Lines 2nd dimension should be of size 4 e.g. [x_start, y_start, x_end, y_end] got {lines.shape[1]}."
            )

    # checks for scores
    if not isinstance(scores, np.ndarray):
        raise ValueError(f"Scores should be numpy array, got {type(scores)}.")

    if len(scores) != 0:
        if len(scores.shape) != 1:
            raise ValueError(
                f"Scores should be of shape (N,) meaning, got {scores.shape}."
            )

        for score in scores:
            if not isinstance(score, (float, np.floating)):
                raise ValueError(f"Scores should be of type float, got {type(score)}.")

    if scores.shape[0] != lines.shape[0]:
        raise ValueError(
            f"Scores should have same length as lines, got {scores.shape[0]} and {lines.shape[0]}."
        )

    line_detections = []
    for i, line in enumerate(lines):
        line_detection = Line()
        line_detection.start_point = dai.Point2f(line[0], line[1])
        line_detection.end_point = dai.Point2f(line[2], line[3])
        line_detection.confidence = float(scores[i])
        line_detections.append(line_detection)

    lines_msg = Lines()
    lines_msg.lines = line_detections
    return lines_msg



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages.creators.segmentation import create_segmentation_message


def test_empty_array():
    with pytest.raises(ValueError, match="Expected 3D input, got 1D input."):
        create_segmentation_message(np.array([]))


def test_wrong_instance():
    with pytest.raises(ValueError, match="Expected numpy array, got <class 'int'>."):
        create_segmentation_message(1)


def test_wrong_dimension():
    with pytest.raises(
        ValueError, match="Expected 1 channel in the third dimension, got 3 channels."
    ):
        create_segmentation_message(np.random.rand(10, 10, 3))


def test_float_array():
    with pytest.raises(
        ValueError, match="Expected int type, got <class 'numpy.float64'>."
    ):
        create_segmentation_message(np.random.rand(10, 10, 1))


def test_complete_types():
    x = (np.random.rand(10, 10, 1) * 255).astype(np.uint8)
    message = create_segmentation_message(x)

    assert isinstance(message, dai.ImgFrame)
    assert message.getWidth() == x.shape[1]
    assert message.getHeight() == x.shape[0]
    assert message.getType() == dai.ImgFrame.Type.RAW8
    assert np.all(np.isclose(message.getFrame(), x[:, :, 0]))


if __name__ == "__main__":
    pytest.main()



from .age_gender import AgeGenderParser
from .classification import ClassificationParser
from .fastsam import FastSAMParser
from .hrnet import HRNetParser
from .image_output import ImageOutputParser
from .keypoints import KeypointParser
from .mediapipe_hand_landmarker import MPHandLandmarkParser
from .mediapipe_palm_detection import MPPalmDetectionParser
from .mlsd import MLSDParser
from .monocular_depth import MonocularDepthParser
from .scrfd import SCRFDParser
from .segmentation import SegmentationParser
from .superanimal_landmarker import SuperAnimalParser
from .thermal_image import ThermalImageParser
from .xfeat import XFeatParser
from .yolo import YOLOExtendedParser
from .yunet import YuNetParser

__all__ = [
    "ImageOutputParser",
    "MonocularDepthParser",
    "YuNetParser",
    "MPPalmDetectionParser",
    "MPHandLandmarkParser",
    "SCRFDParser",
    "SegmentationParser",
    "SuperAnimalParser",
    "KeypointParser",
    "MLSDParser",
    "XFeatParser",
    "ThermalImageParser",
    "ClassificationParser",
    "YOLOExtendedParser",
    "FastSAMParser",
    "AgeGenderParser",
    "HRNetParser",
]



import re

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
)
from depthai_nodes.ml.messages.creators.detection import create_detection_message


def test_not_numpy_array():
    with pytest.raises(
        ValueError, match="Bounding boxes should be numpy array, got <class 'list'>."
    ):
        create_detection_message([1, 2, 3], [0.1, 0.2, 0.3], None, None)


def test_bbox_dim():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Bounding boxes should be of shape (N,4) meaning [...,[x_min, y_min, x_max, y_max],...], got (1,)."
        ),
    ):
        create_detection_message(np.array([1.0]), [0.1], None, None)


def test_bbox_dim_4():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Bounding boxes 2nd dimension should be of size 4 e.g. [x_min, y_min, x_max, y_max] got 3."
        ),
    ):
        create_detection_message(np.array([[1.0, 2.0, 3.0]]), [0.1], None, None)


def test_bbox_valid():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Bounding boxes should be in format [x_min, y_min, x_max, y_max] where xmin < xmax and ymin < ymax."
        ),
    ):
        create_detection_message(np.array([[3.0, 2.0, 1.0, 4.0]]), [0.1], None, None)


def test_scores_not_numpy_array():
    with pytest.raises(
        ValueError, match="Scores should be numpy array, got <class 'list'>."
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0]]), [0.1, 0.2, 0.3], None, None
        )


def test_scores_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Scores should be of shape (N,) meaning, got (1, 1)."),
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0]]), np.array([[0.1]]), None, None
        )


def test_scores_length():
    with pytest.raises(
        ValueError,
        match=re.escape("Scores should have same length as bboxes, got 1 and 2."),
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]]),
            np.array([0.1]),
            None,
            None,
        )


def test_labels_list():
    with pytest.raises(ValueError, match="Labels should be list, got <class 'int'>."):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0]]), np.array([0.1]), 1, None
        )


def test_labels_bbox_lengths():
    with pytest.raises(
        ValueError, match="Labels should have same length as bboxes, got 1 and 2."
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]]),
            np.array([0.1, 0.3]),
            [1],
            None,
        )


def test_in_label_elements():
    with pytest.raises(
        ValueError, match="Labels should be list of integers, got <class 'str'>."
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0]]), np.array([0.1]), ["1"], None
        )


def test_keypoints_list():
    with pytest.raises(
        ValueError, match="Keypoints should be list, got <class 'int'>."
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0]]), np.array([0.1]), None, 1
        )


def test_keypoints_bbox_lengths():
    with pytest.raises(
        ValueError, match="Keypoints should have same length as bboxes, got 1 and 2."
    ):
        create_detection_message(
            np.array([[1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]]),
            np.array([0.1, 0.3]),
            None,
            [[(0, 0), (1, 1)]],
        )


def test_only_bboxes_scores():
    bboxes = np.array(
        [[20.0, 20.0, 40.0, 40.0], [50.0, 50.0, 100.0, 100.0], [10.0, 10.0, 20.0, 20.0]]
    )
    scores = np.array([0.1, 0.2, 0.3])

    message = create_detection_message(bboxes, scores, None, None)

    assert isinstance(message, dai.ImgDetections)
    assert all(
        isinstance(detection, dai.ImgDetection) for detection in message.detections
    )
    assert all(detection.label == 0 for detection in message.detections)
    for i, detection in enumerate(message.detections):
        assert detection.xmin == bboxes[i, 0]
        assert detection.ymin == bboxes[i, 1]
        assert detection.xmax == bboxes[i, 2]
        assert detection.ymax == bboxes[i, 3]
        assert np.isclose(detection.confidence, scores[i])


def test_bboxes_scores_labels():
    bboxes = np.array(
        [[20.0, 20.0, 40.0, 40.0], [50.0, 50.0, 100.0, 100.0], [10.0, 10.0, 20.0, 20.0]]
    )
    scores = np.array([0.1, 0.2, 0.3])
    labels = [1, 2, 3]

    message = create_detection_message(bboxes, scores, labels, None)

    for i, label in enumerate(labels):
        assert message.detections[i].label == label


def test_bboxes_scores_keypoints():
    bboxes = np.array(
        [[20.0, 20.0, 40.0, 40.0], [50.0, 50.0, 100.0, 100.0], [10.0, 10.0, 20.0, 20.0]]
    )
    scores = np.array([0.1, 0.2, 0.3])
    keypoints = [[(0, 0), (1, 1)], [], [(4, 4), (5, 5), (6, 6)]]

    message = create_detection_message(bboxes, scores, None, keypoints)

    assert isinstance(message, ImgDetectionsExtended)
    assert all(
        isinstance(detection, ImgDetectionExtended) for detection in message.detections
    )

    for i, detection in enumerate(message.detections):
        assert detection.keypoints == keypoints[i]
        assert detection.xmin == bboxes[i, 0]
        assert detection.ymin == bboxes[i, 1]
        assert detection.xmax == bboxes[i, 2]
        assert detection.ymax == bboxes[i, 3]
        assert np.isclose(detection.confidence, scores[i])
        assert detection.label == 0


if __name__ == "__main__":
    pytest.main()



import depthai as dai
import numpy as np

from ..messages.creators import create_line_detection_message
from .utils.mlsd import decode_scores_and_points, get_lines


class MLSDParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the M-LSD line detection model. The parser
    is specifically designed to parse the output of the M-LSD model. As the result, the
    node sends out the detected lines in the form of a message.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    topk_n : int
        Number of top candidates to keep.
    score_thr : float
        Confidence score threshold for detected lines.
    dist_thr : float
        Distance threshold for merging lines.

    Output Message/s
    ----------------
    **Type**: LineDetections

    **Description**: LineDetections message containing detected lines and confidence scores.
    """

    def __init__(
        self,
        topk_n=200,
        score_thr=0.10,
        dist_thr=20.0,
    ):
        """Initializes the MLSDParser node.

        @param topk_n: Number of top candidates to keep.
        @type topk_n: int
        @param score_thr: Confidence score threshold for detected lines.
        @type score_thr: float
        @param dist_thr: Distance threshold for merging lines.
        @type dist_thr: float
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()
        self.topk_n = topk_n
        self.score_thr = score_thr
        self.dist_thr = dist_thr

    def setTopK(self, topk_n):
        """Sets the number of top candidates to keep.

        @param topk_n: Number of top candidates to keep.
        @type topk_n: int
        """
        self.topk_n = topk_n

    def setScoreThreshold(self, score_thr):
        """Sets the confidence score threshold for detected lines.

        @param score_thr: Confidence score threshold for detected lines.
        @type score_thr: float
        """
        self.score_thr = score_thr

    def setDistanceThreshold(self, dist_thr):
        """Sets the distance threshold for merging lines.

        @param dist_thr: Distance threshold for merging lines.
        @type dist_thr: float
        """
        self.dist_thr = dist_thr

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            tpMap = output.getTensor("tpMap", dequantize=True).astype(np.float32)
            heat_np = output.getTensor("heat", dequantize=True).astype(np.float32)

            if len(tpMap.shape) != 4:
                raise ValueError("Invalid shape of the tpMap tensor. Should be 4D.")
            if tpMap.shape[3] == 9:
                # We have NWHC format, transform to NCHW
                tpMap = np.transpose(tpMap, (0, 3, 1, 2))

            pts, pts_score, vmap = decode_scores_and_points(tpMap, heat_np, self.topk_n)
            lines, scores = get_lines(
                pts, pts_score, vmap, self.score_thr, self.dist_thr
            )

            message = create_line_detection_message(lines, np.array(scores))
            message.setTimestamp(output.getTimestamp())
            self.out.send(message)



from typing import List

import depthai as dai


class Keypoints(dai.Buffer):
    """Keypoints class for storing keypoints.

    Attributes
    ----------
    keypoints: List[dai.Point3f]
        List of dai.Point3f, each representing a keypoint.
    """

    def __init__(self):
        """Initializes the Keypoints object."""
        super().__init__()
        self._keypoints: List[dai.Point3f] = []

    @property
    def keypoints(self) -> List[dai.Point3f]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: List[dai.Point3f]
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(self, value: List[dai.Point3f]):
        """Sets the keypoints.

        @param value: List of keypoints.
        @type value: List[dai.Point3f]
        @raise TypeError: If the keypoints are not a list.
        @raise TypeError: If each keypoint is not of type dai.Point3f.
        """
        if not isinstance(value, list):
            raise TypeError("keypoints must be a list.")
        for item in value:
            if not isinstance(item, dai.Point3f):
                raise TypeError("All items in keypoints must be of type dai.Point3f.")
        self._keypoints = value


class HandKeypoints(Keypoints):
    """HandKeypoints class for storing hand keypoints.

    Attributes
    ----------
    confidence: float
        Confidence of the hand keypoints.
    handdedness: float
        Handedness of the hand keypoints. 0.0 for left hand and 1.0 for right hand.
    """

    def __init__(self):
        """Initializes the HandKeypoints object."""
        Keypoints.__init__(self)
        self._confidence: float = 0.0
        self._handedness: float = 0.0

    @property
    def confidence(self) -> float:
        """Returns the confidence of the hand keypoints.

        @return: Confidence of the hand keypoints.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the hand keypoints.

        @param value: Confidence of the hand keypoints.
        @type value: float
        @raise TypeError: If the confidence is not a float.
        """
        if not isinstance(value, float):
            raise TypeError("confidence must be a float.")
        self._confidence = value

    @property
    def handedness(self) -> float:
        """Returns the handedness of the hand keypoints.

        @return: Handedness of the hand keypoints.
        @rtype: float
        """
        return self._handedness

    @handedness.setter
    def handedness(self, value: float):
        """Sets the handedness of the hand keypoints.

        @param value: Handedness of the hand keypoints.
        @type value: float
        @raise TypeError: If the handedness is not a float.
        """
        if not isinstance(value, float):
            raise TypeError("handedness must be a float.")
        self._handedness = value



import numpy as np
import pytest

from depthai_nodes.ml.messages import AgeGender
from depthai_nodes.ml.messages.creators.misc import create_age_gender_message


def test_wrong_age():
    with pytest.raises(ValueError, match="Age should be float, got <class 'int'>."):
        create_age_gender_message(1, [0.5, 0.5])


def test_gender_wrong():
    with pytest.raises(
        ValueError, match="Gender_prob should be list, got <class 'str'>."
    ):
        create_age_gender_message(1.0, "female")


def test_gender_empty():
    with pytest.raises(
        ValueError, match="Gender_prob list should have 2 values, got 0"
    ):
        create_age_gender_message(1.0, [])


def test_gender_types():
    with pytest.raises(
        ValueError,
        match="Gender_prob list values must be of type float, instead got <class 'str'>.",
    ):
        create_age_gender_message(1.0, [0.5, "0.5"])


def test_is_probability():
    with pytest.raises(
        ValueError,
        match="Gender_prob list must contain probabilities and sum to 1, got sum 3.7.",
    ):
        create_age_gender_message(1.0, [1.2, 2.5])


def test_correct_types():
    age = 32.4
    gender = [0.35, 0.65]
    message = create_age_gender_message(age, gender)

    assert isinstance(message, AgeGender)
    assert message.age == age
    assert message.gender.classes == ["female", "male"]
    assert np.all(np.isclose(message.gender.scores, gender))


if __name__ == "__main__":
    pytest.main()



import re

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages.creators.depth import create_depth_message

UINT16_MAX_VALUE = 65535
np.random.seed(0)


def test_not_numpy_array():
    with pytest.raises(ValueError, match="Expected numpy array, got <class 'list'>."):
        create_depth_message([1, 2, 3], "relative")


def test_wrong_literal_type():
    with pytest.raises(ValueError):
        create_depth_message(np.array([1, 2, 3]), "wrong")


def test_not_3d_input():
    with pytest.raises(ValueError, match="Expected 2D or 3D input, got 1D input."):
        create_depth_message(np.array([1, 2, 3]), "relative")


def test_wrong_input_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Unexpected image shape. Expected NHW or HWN, got (3, 1, 3)."),
    ):
        create_depth_message(
            np.array([[[1, 2, 3]], [[4, 5, 6]], [[7, 8, 9]]]), "relative"
        )


def test_depth_limit_for_relative_depth():
    depth_map = np.random.rand(320, 640, 1)
    with pytest.raises(
        ValueError,
        match="Invalid depth limit: 1.0. For relative depth, depth limit must be equal to 0.",
    ):
        create_depth_message(depth_map, "relative", 1.0)


def test_no_depth_limit_for_metric_depth():
    depth_map = np.random.rand(320, 640, 1)
    with pytest.raises(
        ValueError,
        match="Invalid depth limit: 0.0. For metric depth, depth limit must be bigger than 0.",
    ):
        create_depth_message(depth_map, "metric")


def test_negative_depth_limit():
    depth_map = np.random.rand(320, 640, 1)
    with pytest.raises(
        ValueError,
        match="Invalid depth limit: -1.0. Depth limit must be bigger than 0.",
    ):
        create_depth_message(depth_map, "metric", -1.0)


def test_relative_depth_map():
    depth_map = np.random.rand(320, 640, 1)

    message = create_depth_message(depth_map, "relative")
    depth_map = depth_map[:, :, 0]

    assert isinstance(message, dai.ImgFrame)
    assert message.getType() == dai.ImgFrame.Type.RAW16
    assert message.getWidth() == 640
    assert message.getHeight() == 320

    frame = message.getFrame()
    assert frame.shape == depth_map.shape
    scaled_depth_map = (
        (depth_map - depth_map.min())
        / (depth_map.max() - depth_map.min())
        * UINT16_MAX_VALUE
    )
    scaled_depth_map = scaled_depth_map.astype(np.uint16)
    assert np.all(np.isclose(frame, scaled_depth_map))


def test_metric_depth_map():
    depth_map = np.random.rand(320, 640, 1)
    depth_limit = 10.0

    message = create_depth_message(depth_map, "metric", depth_limit)
    depth_map = depth_map[:, :, 0]
    depth_map = np.clip(depth_map, a_min=None, a_max=depth_limit)

    assert isinstance(message, dai.ImgFrame)
    assert message.getType() == dai.ImgFrame.Type.RAW16
    assert message.getWidth() == 640
    assert message.getHeight() == 320

    frame = message.getFrame()
    assert frame.shape == depth_map.shape
    scaled_depth_map = (depth_map - 0) / depth_limit * UINT16_MAX_VALUE
    scaled_depth_map = scaled_depth_map.astype(np.uint16)
    assert np.all(np.isclose(frame, scaled_depth_map))


def test_same_depth():
    depth_map = np.ones((320, 640, 1))
    message = create_depth_message(depth_map, "relative")

    assert isinstance(message, dai.ImgFrame)
    assert message.getType() == dai.ImgFrame.Type.RAW16
    assert message.getWidth() == 640
    assert message.getHeight() == 320
    frame = message.getFrame()
    assert np.all(np.isclose(frame, np.zeros((320, 640))))


if __name__ == "__main__":
    pytest.main()



import math

import cv2
import depthai as dai
import numpy as np

from ..messages.creators import create_detection_message
from .utils import decode_detections


class YuNetParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the YuNet face detection model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    conf_threshold : float
        Confidence score threshold for detected faces.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.

    Output Message/s
    ----------------
    **Type**: ImgDetectionsExtended

    **Description**: Message containing bounding boxes, labels, confidence scores, and keypoints of detected faces.
    """

    def __init__(
        self,
        conf_threshold=0.6,
        iou_threshold=0.3,
        max_det=5000,
    ):
        """Initializes the YuNetParser node.

        @param conf_threshold: Confidence score threshold for detected faces.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det

    def setConfidenceThreshold(self, threshold):
        """Sets the confidence score threshold for detected faces.

        @param threshold: Confidence score threshold for detected faces.
        @type threshold: float
        """
        self.conf_threshold = threshold

    def setIOUThreshold(self, threshold):
        """Sets the non-maximum suppression threshold.

        @param threshold: Non-maximum suppression threshold.
        @type threshold: float
        """
        self.iou_threshold = threshold

    def setMaxDetections(self, max_det):
        """Sets the maximum number of detections to keep.

        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        self.max_det = max_det

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            # get strides
            strides = list(
                set(
                    [
                        int(layer_name.split("_")[1])
                        for layer_name in output.getAllLayerNames()
                        if layer_name.startswith(("cls", "obj", "bbox", "kps"))
                    ]
                )
            )

            # get input_size
            stride0 = strides[0]
            cls_stride0_shape = output.getTensor(
                f"cls_{stride0}", dequantize=True
            ).shape
            if len(cls_stride0_shape) == 3:
                _, spatial_positions0, _ = cls_stride0_shape
            elif len(cls_stride0_shape) == 2:
                spatial_positions0, _ = cls_stride0_shape
            input_width = input_height = int(
                math.sqrt(spatial_positions0) * stride0
            )  # TODO: We assume a square input size. How to get input size when height and width are not equal?
            input_size = (input_width, input_height)

            detections = []
            for stride in strides:
                cls = output.getTensor(f"cls_{stride}", dequantize=True)
                cls = cls.astype(np.float32)
                cls = cls.squeeze(0) if cls.shape[0] == 1 else cls

                obj = output.getTensor(f"obj_{stride}", dequantize=True).flatten()
                obj = obj.astype(np.float32)

                bbox = output.getTensor(f"bbox_{stride}", dequantize=True)
                bbox = bbox.astype(np.float32)
                bbox = bbox.squeeze(0) if bbox.shape[0] == 1 else bbox

                kps = output.getTensor(f"kps_{stride}", dequantize=True)
                kps = kps.astype(np.float32)
                kps = kps.squeeze(0) if kps.shape[0] == 1 else kps

                detections += decode_detections(
                    input_size,
                    stride,
                    self.conf_threshold,
                    cls,
                    obj,
                    bbox,
                    kps,
                )

            # non-maximum suppression
            detection_boxes = [detection["bbox"] for detection in detections]
            detection_scores = [detection["score"] for detection in detections]
            indices = cv2.dnn.NMSBoxes(
                detection_boxes,
                detection_scores,
                self.conf_threshold,
                self.iou_threshold,
                top_k=self.max_det,
            )
            detections = np.array(detections)[indices]

            bboxes = []
            for detection in detections:
                xmin, ymin, width, height = detection["bbox"]
                bboxes.append([xmin, ymin, xmin + width, ymin + height])
            scores = [detection["score"] for detection in detections]
            labels = [detection["label"] for detection in detections]
            keypoints = [detection["keypoints"] for detection in detections]

            detections_message = create_detection_message(
                np.array(bboxes),
                np.array(scores),
                labels,
                keypoints,
            )
            detections_message.setTimestamp(output.getTimestamp())

            self.out.send(detections_message)



import depthai as dai
import numpy as np

from ..messages.creators import create_keypoints_message
from .utils.superanimal import get_pose_prediction


class SuperAnimalParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the SuperAnimal landmark model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    score_threshold : float
        Confidence score threshold for detected keypoints.
    scale_factor : float
        Scale factor to divide the keypoints by.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing detected keypoints that exceeds confidence threshold.
    """

    def __init__(
        self,
        score_threshold=0.5,
        scale_factor=256,
    ):
        """Initializes the SuperAnimalParser node.

        @param score_threshold: Confidence score threshold for detected keypoints.
        @type score_threshold: float
        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.score_threshold = score_threshold
        self.scale_factor = scale_factor

    def setScoreThreshold(self, threshold):
        """Sets the confidence score threshold for detected keypoints.

        @param threshold: Confidence score threshold for detected keypoints.
        @type threshold: float
        """
        self.score_threshold = threshold

    def setScaleFactor(self, scale_factor):
        """Sets the scale factor to divide the keypoints by.

        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        """
        self.scale_factor = scale_factor

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            heatmaps = output.getTensor("heatmaps", dequantize=True).astype(np.float32)

            if len(heatmaps.shape) == 3:
                heatmaps = heatmaps.reshape((1,) + heatmaps.shape)

            heatmaps_scale_factor = (
                self.scale_factor / heatmaps.shape[1],
                self.scale_factor / heatmaps.shape[2],
            )

            keypoints = get_pose_prediction(heatmaps, None, heatmaps_scale_factor)[0][0]
            scores = keypoints[:, 2]
            keypoints = keypoints[:, :2] / self.scale_factor

            msg = create_keypoints_message(keypoints, scores, self.score_threshold)
            msg.setTimestamp(output.getTimestamp())

            self.out.send(msg)



import depthai as dai
import numpy as np

from ..messages.creators import create_keypoints_message


class KeypointParser(dai.node.ThreadedHostNode):
    """Parser class for 2D or 3D keypoints models. It expects one ouput layer containing
    keypoints. The number of keypoints must be specified. Moreover, the keypoints are
    normalized by a scale factor if provided.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    scale_factor : float
        Scale factor to divide the keypoints by.
    n_keypoints : int
        Number of keypoints the model detects.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing 2D or 3D keypoints.

    Error Handling
    --------------
    **ValueError**: If the number of keypoints is not specified.

    **ValueError**: If the number of coordinates per keypoint is not 2 or 3.

    **ValueError**: If the number of output layers is not 1.
    """

    def __init__(
        self,
        scale_factor=1,
        n_keypoints=None,
    ):
        """Initializes KeypointParser node.

        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.scale_factor = scale_factor
        self.n_keypoints = n_keypoints

    def setScaleFactor(self, scale_factor):
        """Sets the scale factor to divide the keypoints by.

        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        """
        self.scale_factor = scale_factor

    def setNumKeypoints(self, n_keypoints):
        """Sets the number of keypoints.

        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        """
        self.n_keypoints = n_keypoints

    def run(self):
        if self.n_keypoints is None:
            raise ValueError("Number of keypoints must be specified!")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()

            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )

            keypoints = output.getTensor(output_layer_names[0], dequantize=True).astype(
                np.float32
            )
            num_coords = int(np.prod(keypoints.shape) / self.n_keypoints)

            if num_coords not in [2, 3]:
                raise ValueError(
                    f"Expected 2 or 3 coordinates per keypoint, got {num_coords}."
                )

            keypoints = keypoints.reshape(self.n_keypoints, num_coords)

            keypoints /= self.scale_factor

            msg = create_keypoints_message(keypoints)
            msg.setTimestamp(output.getTimestamp())

            self.out.send(msg)



import depthai as dai

from ..messages import Classifications


class AgeGender(dai.Buffer):
    def __init__(self):
        super().__init__()
        self._age: float = None
        self._gender = Classifications()

    @property
    def age(self) -> float:
        return self._age

    @age.setter
    def age(self, value: float):
        if not isinstance(value, float):
            raise TypeError(
                f"start_point must be of type float, instead got {type(value)}."
            )
        self._age = value

    @property
    def gender(self) -> Classifications:
        return self._gender

    @gender.setter
    def gender(self, value: Classifications):
        if not isinstance(value, Classifications):
            raise TypeError(
                f"gender must be of type Classifications, instead got {type(value)}."
            )
        self._gender = value



from typing import List

import depthai as dai


class Classifications(dai.Buffer):
    """Classification class for storing the class names and their respective scores.

    Attributes
    ----------
    classes : list[str]
        A list of classes.
    scores : list[float]
        A list of corresponding probability scores.
    """

    def __init__(self):
        """Initializes the Classifications object and sets the classes and scores to
        empty lists."""
        dai.Buffer.__init__(self)
        self._classes: List[str] = []
        self._scores: List[float] = []

    @property
    def classes(self) -> List:
        """Returns the list of classes."""
        return self._classes

    @property
    def scores(self) -> List:
        """Returns the list of scores."""
        return self._scores

    @classes.setter
    def classes(self, class_names: List[str]):
        """Sets the list of classes.

        @param classes: A list of class names.
        """
        self._classes = class_names

    @scores.setter
    def scores(self, scores: List[float]):
        """Sets the list of scores.

        @param scores: A list of scores.
        """
        self._scores = scores



import cv2
import depthai as dai
import numpy as np

from depthai_nodes.ml.messages import ImgDetectionsExtended, Lines

from .colors import get_yolo_colors
from .messages import (
    parse_detection_message,
    parse_line_detection_message,
    parse_yolo_kpts_message,
)


def visualize_detections(
    frame: dai.ImgFrame, message: dai.ImgDetections, extraParams: dict
):
    """Visualizes the detections on the frame.

    Also, checks if there are any keypoints available to visualize.
    """
    labels = extraParams.get("classes", None)
    detections = parse_detection_message(message)
    for detection in detections:
        xmin, ymin, xmax, ymax = (
            detection.xmin,
            detection.ymin,
            detection.xmax,
            detection.ymax,
        )
        if xmin > 1 or ymin > 1 or xmax > 1 or ymax > 1:
            xmin = int(xmin)
            ymin = int(ymin)
            xmax = int(xmax)
            ymax = int(ymax)
        else:
            xmin = int(xmin * frame.shape[1])
            ymin = int(ymin * frame.shape[0])
            xmax = int(xmax * frame.shape[1])
            ymax = int(ymax * frame.shape[0])
        cv2.rectangle(
            frame, (int(xmin), int(ymin)), (int(xmax), int(ymax)), (255, 0, 0), 2
        )

        try:
            keypoints = detection.keypoints
            for kp in keypoints:
                cv2.circle(
                    frame,
                    (int(kp[0] * frame.shape[1]), int(kp[1] * frame.shape[0])),
                    5,
                    (0, 0, 255),
                    -1,
                )
        except Exception:
            pass

        cv2.putText(
            frame,
            f"{detection.confidence * 100:.2f}%",
            (int(xmin) + 10, int(ymin) + 20),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )
        if labels is not None:
            cv2.putText(
                frame,
                labels[detection.label],
                (int(xmin) + 10, int(ymin) + 40),
                cv2.FONT_HERSHEY_TRIPLEX,
                0.5,
                255,
            )

    cv2.imshow("Detections", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_line_detections(frame: dai.ImgFrame, message: Lines, extraParams: dict):
    """Visualizes the lines on the frame."""
    lines = parse_line_detection_message(message)
    h, w = frame.shape[:2]
    for line in lines:
        x1 = line.start_point.x * w
        y1 = line.start_point.y * h
        x2 = line.end_point.x * w
        y2 = line.end_point.y * h
        cv2.line(frame, (int(x1), int(y1)), (int(x2), int(y2)), (0, 255, 0), 3, 16)

    cv2.putText(
        frame,
        f"Number of lines: {len(lines)}",
        (2, frame.shape[0] - 4),
        cv2.FONT_HERSHEY_COMPLEX,
        0.5,
        (255, 0, 0),
    )
    cv2.imshow("Lines", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False


def visualize_yolo_extended(
    frame: dai.ImgFrame, message: ImgDetectionsExtended, extraParams: dict
):
    """Visualizes the YOLO pose detections or instance segmentation on the frame."""
    detections = parse_yolo_kpts_message(message)

    colors = get_yolo_colors()
    classes = extraParams.get("classes", None)
    if classes is None:
        raise ValueError("Classes are required for visualization.")
    task = extraParams.get("n_keypoints", None)
    if task is None:
        task = "segmentation"
    else:
        task = "keypoints"

    overlay = np.zeros_like(frame)

    for detection in detections:
        xmin, ymin, xmax, ymax = (
            detection.xmin,
            detection.ymin,
            detection.xmax,
            detection.ymax,
        )
        cv2.rectangle(
            frame, (int(xmin), int(ymin)), (int(xmax), int(ymax)), (255, 0, 0), 2
        )
        cv2.putText(
            frame,
            f"{detection.confidence * 100:.2f}%",
            (int(xmin) + 10, int(ymin) + 20),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )
        cv2.putText(
            frame,
            f"{classes[detection.label]}",
            (int(xmin) + 10, int(ymin) + 40),
            cv2.FONT_HERSHEY_TRIPLEX,
            0.5,
            255,
        )

        if task == "keypoints":
            keypoints = detection.keypoints
            for keypoint in keypoints:
                x, y, visibility = keypoint[0], keypoint[1], keypoint[2]
                if visibility > 0.8:
                    cv2.circle(frame, (int(x), int(y)), 2, (0, 255, 0), -1)
        else:
            mask = detection.mask
            mask = cv2.resize(mask, (frame.shape[1], frame.shape[0]))
            mask = detection.mask
            mask = cv2.resize(mask, (512, 288))
            label = detection.label
            overlay[mask > 0] = colors[(label % len(colors))]

    if task == "segmentation":
        frame = cv2.addWeighted(overlay, 0.8, frame, 0.5, 0, None)
    cv2.imshow("YOLO Pose Estimation", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



from .classification import Classifications
from .img_detections import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
)
from .keypoints import HandKeypoints, Keypoints
from .lines import Line, Lines
from .misc import AgeGender
from .segmentation import SegmentationMasks

__all__ = [
    "ImgDetectionExtended",
    "ImgDetectionsExtended",
    "HandKeypoints",
    "Keypoints",
    "Line",
    "Lines",
    "Classifications",
    "SegmentationMasks",
    "AgeGender",
]



import argparse
from typing import Tuple


def initialize_argparser():
    """Initialize the argument parser for the script."""
    parser = argparse.ArgumentParser()
    parser.description = "General example script to run any model available in HubAI on DepthAI device. \
        All you need is a model slug of the model and the script will download the model from HubAI and create \
        the whole pipeline with visualizations. You also need a DepthAI device connected to your computer. \
        Currently, only RVC2 devices are supported."

    parser.add_argument(
        "-s",
        "--model_slug",
        help="slug of the model in HubAI.",
        required=True,
        type=str,
    )

    args = parser.parse_args()

    return parser, args


def parse_model_slug(args: argparse.Namespace) -> Tuple[str, str]:
    """Parse the model slug from the arguments.

    Returns the model slug and model version slug.
    """
    model_slug = args.model_slug

    # parse the model slug
    if ":" not in model_slug:
        raise NameError(
            "Please provide the model slug in the format of 'model_slug:model_version_slug'"
        )

    model_slug_parts = model_slug.split(":")
    model_slug = model_slug_parts[0]
    model_version_slug = model_slug_parts[1]

    return model_slug, model_version_slug



import re

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages.creators.thermal import create_thermal_message


def test_dimension():
    with pytest.raises(ValueError):
        create_thermal_message(np.array([[0], [1], [1]]))


def test_shape():
    with pytest.raises(
        ValueError,
        match=re.escape("Unexpected image shape. Expected CHW or HWC, got (2, 2, 2)."),
    ):
        create_thermal_message(np.array([[[0, 1], [1, 1]], [[0, 1], [1, 1]]]))


def test_float():
    with pytest.raises(ValueError, match="Expected integer values, got float."):
        create_thermal_message(np.array([[[0.3, 0.17], [0.5, 0.1]]]))


def test_negative():
    with pytest.raises(
        ValueError, match="All values of thermal_image have to be non-negative."
    ):
        create_thermal_message(np.array([[[0, 1], [1, -1]]]))


def test_return():
    image = np.array([[[5, 7, 8], [4, 10, 5]]])
    imgFrame = create_thermal_message(image)
    assert imgFrame.getType() == dai.ImgFrame.Type.RAW16
    assert imgFrame.getWidth() == 3
    assert imgFrame.getHeight() == 2
    assert np.all(np.isclose(imgFrame.getFrame(), image))


if __name__ == "__main__":
    pytest.main()



from typing import Tuple

import depthai as dai
import numpy as np

from ..messages.creators import create_tracked_features_message
from .utils.xfeat import detect_and_compute, match


class XFeatParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the XFeat model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    original_size : Tuple[float, float]
        Original image size.
    input_size : Tuple[float, float]
        Input image size.
    max_keypoints : int
        Maximum number of keypoints to keep.
    previous_results : np.ndarray
        Previous results from the model. Previous results are used to match keypoints between two frames.

    Output Message/s
    ----------------
    **Type**: dai.TrackedFeatures

    **Description**: TrackedFeatures message containing matched keypoints with the same ID.

    Error Handling
    --------------
    **ValueError**: If the original image size is not specified.
    """

    def __init__(
        self,
        original_size: Tuple[float, float] = None,
        input_size: Tuple[float, float] = (640, 352),
        max_keypoints: int = 4096,
    ):
        """Initializes the XFeatParser node.

        @param original_size: Original image size.
        @type original_size: Tuple[float, float]
        @param input_size: Input image size.
        @type input_size: Tuple[float, float]
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()
        self.original_size = original_size
        self.input_size = input_size
        self.max_keypoints = max_keypoints
        self.previous_results = None

    def setOriginalSize(self, original_size):
        """Sets the original image size.

        @param original_size: Original image size.
        @type original_size: Tuple[float, float]
        """
        self.original_size = original_size

    def setInputSize(self, input_size):
        """Sets the input image size.

        @param input_size: Input image size.
        @type input_size: Tuple[float, float]
        """
        self.input_size = input_size

    def setMaxKeypoints(self, max_keypoints):
        """Sets the maximum number of keypoints to keep.

        @param max_keypoints: Maximum number of keypoints.
        @type max_keypoints: int
        """
        self.max_keypoints = max_keypoints

    def run(self):
        if self.original_size is None:
            raise ValueError("Original image size must be specified!")

        resize_rate_w = self.original_size[0] / self.input_size[0]
        resize_rate_h = self.original_size[1] / self.input_size[1]

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            feats = output.getTensor("feats", dequantize=True).astype(np.float32)
            keypoints = output.getTensor("keypoints", dequantize=True).astype(
                np.float32
            )

            if len(feats.shape) == 3:
                feats = feats.reshape((1,) + feats.shape).transpose(0, 3, 1, 2)
            if len(keypoints.shape) == 3:
                keypoints = keypoints.reshape((1,) + keypoints.shape).transpose(
                    0, 3, 1, 2
                )

            result = detect_and_compute(
                feats,
                keypoints,
                resize_rate_w,
                resize_rate_h,
                self.input_size,
                self.max_keypoints,
            )

            if result is not None:
                result = result[0]
            else:
                matched_points = dai.TrackedFeatures()
                matched_points.setTimestamp(output.getTimestamp())
                self.out.send(matched_points)
                continue

            if self.previous_results is not None:
                mkpts0, mkpts1 = match(self.previous_results, result)
                matched_points = create_tracked_features_message(mkpts0, mkpts1)
                matched_points.setTimestamp(output.getTimestamp())
                self.out.send(matched_points)

            # save the result from first frame
            self.previous_results = result



import depthai as dai
import numpy as np

from ..messages.creators import create_detection_message
from .utils.scrfd import decode_scrfd


class SCRFDParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the SCRFD face detection model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    conf_threshold : float
        Confidence score threshold for detected faces.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.
    feat_stride_fpn : tuple
        Tuple of the feature strides.
    num_anchors : int
        Number of anchors.
    input_size : tuple
        Input size of the model.

    Output Message/s
    ----------------
    **Type**: dai.ImgDetections

    **Description**: ImgDetections message containing bounding boxes, labels, and confidence scores of detected faces.
    """

    def __init__(
        self,
        conf_threshold=0.5,
        iou_threshold=0.5,
        max_det=100,
        input_size=(640, 640),
        feat_stride_fpn=(8, 16, 32),
        num_anchors=2,
    ):
        """Initializes the SCRFDParser node.

        @param conf_threshold: Confidence score threshold for detected faces.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        @param feat_stride_fpn: List of the feature strides.
        @type feat_stride_fpn: tuple
        @param num_anchors: Number of anchors.
        @type num_anchors: int
        @param input_size: Input size of the model.
        @type input_size: tuple
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det

        self.feat_stride_fpn = feat_stride_fpn
        self.num_anchors = num_anchors
        self.input_size = input_size

    def setConfidenceThreshold(self, threshold):
        """Sets the confidence score threshold for detected faces.

        @param threshold: Confidence score threshold for detected faces.
        @type threshold: float
        """
        self.conf_threshold = threshold

    def setIOUThreshold(self, threshold):
        """Sets the non-maximum suppression threshold.

        @param threshold: Non-maximum suppression threshold.
        @type threshold: float
        """
        self.iou_threshold = threshold

    def setMaxDetections(self, max_det):
        """Sets the maximum number of detections to keep.

        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        self.max_det = max_det

    def setFeatStrideFPN(self, feat_stride_fpn):
        """Sets the feature stride of the FPN.

        @param feat_stride_fpn: Feature stride of the FPN.
        @type feat_stride_fpn: list
        """
        self.feat_stride_fpn = feat_stride_fpn

    def setInputSize(self, input_size):
        """Sets the input size of the model.

        @param input_size: Input size of the model.
        @type input_size: list
        """
        self.input_size = input_size

    def setNumAnchors(self, num_anchors):
        """Sets the number of anchors.

        @param num_anchors: Number of anchors.
        @type num_anchors: int
        """
        self.num_anchors = num_anchors

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            scores_concatenated = []
            bboxes_concatenated = []
            kps_concatenated = []

            for stride in self.feat_stride_fpn:
                score_layer_name = f"score_{stride}"
                bbox_layer_name = f"bbox_{stride}"
                kps_layer_name = f"kps_{stride}"
                if score_layer_name not in output.getAllLayerNames():
                    raise ValueError(
                        f"Layer {score_layer_name} not found in the model output."
                    )
                if bbox_layer_name not in output.getAllLayerNames():
                    raise ValueError(
                        f"Layer {bbox_layer_name} not found in the model output."
                    )
                if kps_layer_name not in output.getAllLayerNames():
                    raise ValueError(
                        f"Layer {kps_layer_name} not found in the model output."
                    )

                score_tensor = (
                    output.getTensor(score_layer_name, dequantize=True)
                    .flatten()
                    .astype(np.float32)
                )
                bbox_tensor = (
                    output.getTensor(bbox_layer_name, dequantize=True)
                    .reshape(len(score_tensor), 4)
                    .astype(np.float32)
                )
                kps_tensor = (
                    output.getTensor(kps_layer_name, dequantize=True)
                    .reshape(len(score_tensor), 10)
                    .astype(np.float32)
                )

                scores_concatenated.append(score_tensor)
                bboxes_concatenated.append(bbox_tensor)
                kps_concatenated.append(kps_tensor)

            bboxes, scores, keypoints = decode_scrfd(
                bboxes_concatenated=bboxes_concatenated,
                scores_concatenated=scores_concatenated,
                kps_concatenated=kps_concatenated,
                feat_stride_fpn=self.feat_stride_fpn,
                input_size=self.input_size,
                num_anchors=self.num_anchors,
                score_threshold=self.conf_threshold,
                nms_threshold=self.iou_threshold,
            )
            detection_msg = create_detection_message(
                bboxes, scores, None, keypoints.tolist()
            )
            detection_msg.setTimestamp(output.getTimestamp())

            self.out.send(detection_msg)



import depthai as dai
import numpy as np


def create_thermal_message(thermal_image: np.ndarray) -> dai.ImgFrame:
    """Create a DepthAI message for thermal image.

    @param thermal_image: A NumPy array representing the thermal image with shape (CHW
        or HWC).
    @type thermal_image: np.array
    @return: An ImgFrame object containing the thermal information.
    @rtype: dai.ImgFrame
    @raise ValueError: If the input is not a NumPy array.
    @raise ValueError: If the input is not 3D.
    @raise ValueError: If the input shape is not CHW or HWC.
    """

    if not isinstance(thermal_image, np.ndarray):
        raise ValueError(f"Expected numpy array, got {type(thermal_image)}.")
    if len(thermal_image.shape) != 3:
        raise ValueError(f"Expected 3D input, got {len(thermal_image.shape)}D input.")

    if thermal_image.shape[0] == 1:
        thermal_image = thermal_image[0, :, :]  # CHW to HW
    elif thermal_image.shape[2] == 1:
        thermal_image = thermal_image[:, :, 0]  # HWC to HW
    else:
        raise ValueError(
            f"Unexpected image shape. Expected CHW or HWC, got {thermal_image.shape}."
        )

    if isinstance(thermal_image.flat[0], float):
        raise ValueError("Expected integer values, got float.")

    if np.any(thermal_image < 0):
        raise ValueError("All values of thermal_image have to be non-negative.")

    thermal_image = thermal_image.astype(np.uint16)

    imgFrame = dai.ImgFrame()
    imgFrame.setFrame(thermal_image)
    imgFrame.setWidth(thermal_image.shape[1])
    imgFrame.setHeight(thermal_image.shape[0])
    imgFrame.setType(dai.ImgFrame.Type.RAW16)

    return imgFrame



import depthai as dai

from ..messages.creators import create_image_message
from .utils import unnormalize_image


class ImageOutputParser(dai.node.ThreadedHostNode):
    """Parser class for image-to-image models (e.g. DnCNN3, zero-dce etc.) where the
    output is a modifed image (denoised, enhanced etc.).

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    output_is_bgr : bool
        Flag indicating if the output image is in BGR (Blue-Green-Red) format.

    Output Message/s
    -------
    **Type**: dai.ImgFrame

    **Description**: Image message containing the output image e.g. denoised or enhanced images.

    Error Handling
    --------------
    **ValueError**: If the output is not 3- or 4-dimensional.

    **ValueError**: If the number of output layers is not 1.
    """

    def __init__(self, output_is_bgr=False):
        """Initializes ImageOutputParser node.

        @param output_is_bgr: Flag indicating if the output image is in BGR.
        @type output_is_bgr: bool
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.output_is_bgr = output_is_bgr

    def setBGROutput(self):
        """Sets the flag indicating that output image is in BGR."""
        self.output_is_bgr = True

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()
            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )

            output_image = output.getTensor(output_layer_names[0], dequantize=True)

            if len(output_image.shape) == 4:
                image = output_image[0]
            elif len(output_image.shape) == 3:
                image = output_image
            else:
                raise ValueError(
                    f"Expected 3- or 4-dimensional output, got {len(output_image.shape)}-dimensional",
                )

            image = unnormalize_image(image)

            image_message = create_image_message(
                image=image,
                is_bgr=self.output_is_bgr,
            )
            image_message.setTimestamp(output.getTimestamp())

            self.out.send(image_message)



import cv2
import depthai as dai

from depthai_nodes.ml.messages import Keypoints

from .messages import parse_keypoints_message


def visualize_keypoints(frame: dai.ImgFrame, message: Keypoints, extraParams: dict):
    """Visualizes the keypoints on the frame."""
    keypoints = parse_keypoints_message(message)

    for kp in keypoints:
        x = int(kp.x * frame.shape[1])
        y = int(kp.y * frame.shape[0])
        cv2.circle(frame, (x, y), 1, (0, 255, 0), -1)

    cv2.imshow("Keypoints", frame)
    if cv2.waitKey(1) == ord("q"):
        cv2.destroyAllWindows()
        return True

    return False



import depthai as dai
import numpy as np

from ..messages.creators import create_hand_keypoints_message


class MPHandLandmarkParser(dai.node.ThreadedHostNode):
    """Parser class for MediaPipe Hand landmark model. It parses the output of the
    MediaPipe Hand landmark model containing 21 3D hand landmarks. The landmarks are
    normalized and sent as a message to the output. Besides landmarks, the message
    contains confidence score and handedness score (right or left hand).

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    score_threshold : float
        Confidence score threshold for hand landmarks.
    scale_factor : float
        Scale factor to divide the landmarks by.

    Output Message/s
    ----------------
    **Type**: HandLandmarks

    **Description**: HandLandmarks message containing normalized 21 3D landmarks, confidence score, and handedness score (right or left hand).

    See also
    --------
    Official MediaPipe Hands solution:
    https://ai.google.dev/edge/mediapipe/solutions/vision/hand_landmarker
    """

    def __init__(self, score_threshold=0.5, scale_factor=224):
        """Initialize MPHandLandmarkParser node.

        @param score_threshold: Confidence score threshold for hand landmarks.
        @type score_threshold: float
        @param scale_factor: Scale factor to divide the landmarks by.
        @type scale_factor: float
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.score_threshold = score_threshold
        self.scale_factor = scale_factor

    def setScoreThreshold(self, threshold):
        """Set the confidence score threshold for hand landmarks.

        @param threshold: Confidence score threshold for hand landmarks.
        @type threshold: float
        """
        self.score_threshold = threshold

    def setScaleFactor(self, scale_factor):
        """Set the scale factor to divide the landmarks by.

        @param scale_factor: Scale factor to divide the landmarks by.
        @type scale_factor: float
        """
        self.scale_factor = scale_factor

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            landmarks = (
                output.getTensor("Identity", dequantize=True)
                .reshape(21, 3)
                .astype(np.float32)
            )
            hand_score = (
                output.getTensor("Identity_1", dequantize=True)
                .reshape(-1)
                .astype(np.float32)
            )
            handedness = (
                output.getTensor("Identity_2", dequantize=True)
                .reshape(-1)
                .astype(np.float32)
            )
            hand_score = hand_score[0]
            handedness = handedness[0]

            # normalize landmarks
            landmarks /= self.scale_factor

            hand_landmarks_msg = create_hand_keypoints_message(
                landmarks, float(handedness), float(hand_score), self.score_threshold
            )
            hand_landmarks_msg.setTimestamp(output.getTimestamp())
            self.out.send(hand_landmarks_msg)



import depthai as dai
import numpy as np


class SegmentationMasks(dai.Buffer):
    """SegmentationMasks class for storing segmentation masks.

    Attributes
    ----------
    masks: np.ndarray
        Mask coefficients.
    """

    def __init__(self):
        """Initializes the SegmentationMasks object."""
        super().__init__()
        self._masks: np.ndarray = np.array([])

    @property
    def masks(self) -> np.ndarray:
        """Returns the masks coefficients.

        @return: Masks coefficients.
        @rtype: np.ndarray
        """
        return self._masks

    @masks.setter
    def masks(self, values: np.ndarray):
        """Sets the masks coefficients.

        @param value: Masks coefficients.
        @type value: np.ndarray
        @raise TypeError: If the masks is not a numpy array.
        @raise ValueError: If the masks is not 3D.
        """
        if not isinstance(values, np.ndarray):
            raise TypeError("Masks must be a numpy array.")
        if len(values.shape) != 3:
            raise ValueError("Masks must be 3D.")
        self._masks = values



import re

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages import Keypoints
from depthai_nodes.ml.messages.creators.keypoints import create_keypoints_message


def test_keypoints_not_list():
    with pytest.raises(
        ValueError, match="Keypoints should be numpy array or list, got <class 'int'>."
    ):
        create_keypoints_message(1, 0.5, 0.8)


def test_scores_not_list():
    with pytest.raises(
        ValueError, match="Scores should be numpy array or list, got <class 'int'>."
    ):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), 1, 0.5)


def test_none_keypoints():
    with pytest.raises(
        ValueError,
        match="Keypoints should be numpy array or list, got <class 'NoneType'>.",
    ):
        create_keypoints_message(None, 0.5, 0.8)


def test_none_scores():
    create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), None, None)


def test_empty_keypoints():
    create_keypoints_message([], [], 0.8)


def test_keypoints_and_scores_length():
    with pytest.raises(
        ValueError,
        match="Keypoints and scores should have the same length. Got 1 keypoints and 2 scores.",
    ):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), [1.0, 2.0], 0.8)


def test_non_float_scores():
    with pytest.raises(ValueError, match="Scores should only contain float values."):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), ["0.5"], 0.5)


def test_2d_scores():
    with pytest.raises(ValueError, match="Scores should only contain float values."):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), np.array([[0.5]]), 0.5)


def test_confidence_threshold_not_float():
    with pytest.raises(
        ValueError, match="The confidence_threshold should be float, got <class 'str'>."
    ):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), [0.5], "0.8")


def test_confidence_threshold_between_0_and_1():
    with pytest.raises(
        ValueError,
        match="The confidence_threshold should be between 0 and 1, got confidence_threshold 1.5",
    ):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), [0.5], 1.50)


def test_negative_scores():
    with pytest.raises(
        ValueError, match="Scores should only contain values between 0 and 1."
    ):
        create_keypoints_message(np.array([[1.0, 2.0, 3.0]]), [-0.5], 0.5)


def test_1d_keypoint():
    with pytest.raises(
        ValueError, match="All keypoints should be of dimension 2 or 3, got dimension 1"
    ):
        create_keypoints_message(np.array([[1.0]]), [0.5], 0.8)


def test_consistent_keypoint_dims():
    with pytest.raises(
        ValueError,
        match=re.escape(
            "All keypoints have to be of same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
        ),
    ):
        create_keypoints_message(
            [[1.0, 2.0, 3.0], [1.0, 3.0], [1.0, 2.0, 3.0]], [0.5, 0.8, 0.7], 0.5
        )


def test_nonlist_keypoint():
    with pytest.raises(
        ValueError,
        match="Keypoints should be list of lists or np.array, got list of <class 'str'>.",
    ):
        create_keypoints_message(
            [[1.0, 2.0, 3.0], "2", [1.0, 2.0, 3.0]], [0.5, 0.8, 0.7], 0.5
        )


def test_nonfloat_keypoint():
    with pytest.raises(
        ValueError,
        match="Keypoints inner list should contain only float, got <class 'str'>.",
    ):
        create_keypoints_message(
            [[1.0, 2.0, 3.0], [2.0, "3", 4.0], [1.0, 2.0, 3.0]], [0.5, 0.8, 0.7], 0.5
        )


def test_3d_array():
    with pytest.raises(
        ValueError,
        match=re.escape("Keypoints should be of shape (N,2 or 3) got (3, 2, 3)."),
    ):
        create_keypoints_message(
            np.array(
                [
                    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
                    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
                    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
                ]
            ),
            [0.5, 0.8, 0.7],
            0.5,
        )


def test_all_scores_below_thr():
    message = create_keypoints_message([[1.0, 2.0, 3.0]], [0.4], 0.6)

    assert isinstance(message, Keypoints)
    assert message.keypoints == []


def test_all_scores_above_thr():
    message = create_keypoints_message([[1.0, 2.0, 3.0]], [0.8], 0.6)

    true_pt = np.array([1.0, 2.0, 3.0])

    assert len(message.keypoints) == 1
    assert isinstance(message.keypoints[0], dai.Point3f)
    assert message.keypoints[0].x == true_pt[0]
    assert message.keypoints[0].y == true_pt[1]
    assert message.keypoints[0].z == true_pt[2]


def test_2d_keypoints():
    message = create_keypoints_message([[1.0, 2.0], [3.0, 4.0]], [0.8, 0.5], 0.6)

    assert len(message.keypoints) == 1
    assert message.keypoints[0].x == 1.0
    assert message.keypoints[0].y == 2.0
    assert message.keypoints[0].z == 0.0


if __name__ == "__main__":
    pytest.main()



import depthai as dai
import numpy as np

from ..messages.creators import create_keypoints_message


class HRNetParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the HRNet pose estimation model. The code is inspired by https://github.com/ibaiGorordo/ONNX-HRNET-Human-Pose-Estimation.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    score_threshold : float
        Confidence score threshold for detected keypoints.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing detected body keypoints.
    """

    def __init__(self, score_threshold=0.5):
        """Initializes the HRNetParser node.

        @param score_threshold: Confidence score threshold for detected keypoints.
        @type score_threshold: float
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = dai.Node.Input(self)
        self.out = dai.Node.Output(self)

        self.score_threshold = score_threshold

    def setScoreThreshold(self, threshold):
        """Sets the confidence score threshold for the detected body keypoints.

        @param threshold: Confidence score threshold for detected keypoints.
        @type threshold: float
        """
        self.score_threshold = threshold

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            heatmaps = output.getTensor("heatmaps", dequantize=True)

            if len(heatmaps.shape) == 4:
                heatmaps = heatmaps[0]
            if heatmaps.shape[2] == 16:  # HW_ instead of _HW
                heatmaps = heatmaps.transpose(2, 0, 1)
            _, map_h, map_w = heatmaps.shape

            scores = np.array([np.max(heatmap) for heatmap in heatmaps])
            keypoints = np.array(
                [
                    np.unravel_index(heatmap.argmax(), heatmap.shape)
                    for heatmap in heatmaps
                ]
            )
            keypoints = keypoints.astype(np.float32)
            keypoints = keypoints[:, ::-1] / np.array(
                [map_w, map_h]
            )  # normalize keypoints to [0, 1]

            keypoints_message = create_keypoints_message(
                keypoints=keypoints,
                scores=scores,
                confidence_threshold=self.score_threshold,
            )
            keypoints_message.setTimestamp(output.getTimestamp())

            self.out.send(keypoints_message)



import depthai as dai

from ..messages.creators import create_age_gender_message


class AgeGenderParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the Age-Gender regression model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.

    Output Message/s
    ----------------
    **Type**: AgeGender

    **Description**: Message containing the detected person age and Classfications object for storing information about the detected person's gender.
    """

    def __init__(self):
        """Initializes the AgeGenderParser node."""
        dai.node.ThreadedHostNode.__init__(self)
        self.input = dai.Node.Input(self)
        self.out = dai.Node.Output(self)

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            age = output.getTensor("age_conv3", dequantize=True).item()
            age *= 100  # convert to years
            prob = output.getTensor("prob", dequantize=True).flatten().tolist()

            age_gender_message = create_age_gender_message(age=age, gender_prob=prob)
            age_gender_message.setTimestamp(output.getTimestamp())

            self.out.send(age_gender_message)



import depthai as dai

from ..messages.creators import create_thermal_message


class ThermalImageParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of models with thermal image output (e.g.
    UGSR-FA).

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.

    Output Message/s
    ----------------
    **Type**: dai.ImgFrame

    **Description**: Thermal message containing the thermal image.
    """

    def __init__(self):
        """Initializes the ThermalImageParser node."""
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()
            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )
            output = output.getTensor(output_layer_names[0])

            thermal_map = output[0]

            thermal_message = create_thermal_message(thermal_map=thermal_map)
            thermal_message.setTimestamp(output.getTimestamp())
            self.out.send(thermal_message)



import cv2
import depthai as dai
import numpy as np

from ..messages.creators import create_detection_message
from .utils import generate_anchors_and_decode


class MPPalmDetectionParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the Mediapipe Palm detection model. As the
    result, the node sends out the detected hands in the form of a message containing
    bounding boxes, labels, and confidence scores.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.Parser sends the processed network results to this output in form of messages. It is a linking point from which the processed network results are retrieved.
    conf_threshold : float
        Confidence score threshold for detected hands.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.
    scale : int
        Scale of the input image.

    Output Message/s
    -------
    **Type**: dai.ImgDetections

    **Description**: ImgDetections message containing bounding boxes, labels, and confidence scores of detected hands.

    See also
    --------
    Official MediaPipe Hands solution:
    https://ai.google.dev/edge/mediapipe/solutions/vision/hand_landmarker
    """

    def __init__(self, conf_threshold=0.5, iou_threshold=0.5, max_det=100, scale=192):
        """Initializes the MPPalmDetectionParser node.

        @param conf_threshold: Confidence score threshold for detected hands.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det
        self.scale = scale

    def setConfidenceThreshold(self, threshold):
        """Sets the confidence score threshold for detected hands.

        @param threshold: Confidence score threshold for detected hands.
        @type threshold: float
        """
        self.conf_threshold = threshold

    def setIOUThreshold(self, threshold):
        """Sets the non-maximum suppression threshold.

        @param threshold: Non-maximum suppression threshold.
        @type threshold: float
        """
        self.iou_threshold = threshold

    def setMaxDetections(self, max_det):
        """Sets the maximum number of detections to keep.

        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        self.max_det = max_det

    def setScale(self, scale):
        """Sets the scale of the input image.

        @param scale: Scale of the input image.
        @type scale: int
        """
        self.scale = scale

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            all_tensors = output.getAllLayerNames()

            bboxes = None
            scores = None

            for tensor_name in all_tensors:
                tensor = output.getTensor(tensor_name, dequantize=True).astype(
                    np.float32
                )
                if bboxes is None:
                    bboxes = tensor
                    scores = tensor
                else:
                    bboxes = bboxes if tensor.shape[-1] < bboxes.shape[-1] else tensor
                    scores = tensor if tensor.shape[-1] < scores.shape[-1] else scores

            bboxes = bboxes.reshape(-1, 18)
            scores = scores.reshape(-1)

            if bboxes is None or scores is None:
                raise ValueError("No valid output tensors found.")

            decoded_bboxes = generate_anchors_and_decode(
                bboxes=bboxes,
                scores=scores,
                threshold=self.conf_threshold,
                scale=self.scale,
            )

            bboxes = []
            scores = []

            for hand in decoded_bboxes:
                extended_points = hand.rect_points
                xmin = int(min([point[0] for point in extended_points]))
                ymin = int(min([point[1] for point in extended_points]))
                xmax = int(max([point[0] for point in extended_points]))
                ymax = int(max([point[1] for point in extended_points]))

                bboxes.append([xmin, ymin, xmax, ymax])
                scores.append(hand.pd_score)

            indices = cv2.dnn.NMSBoxes(
                bboxes,
                scores,
                self.conf_threshold,
                self.iou_threshold,
                top_k=self.max_det,
            )
            bboxes = np.array(bboxes)[indices]
            scores = np.array(scores)[indices]

            bboxes = bboxes.astype(np.float32) / self.scale

            detections_msg = create_detection_message(bboxes, scores, labels=None)
            detections_msg.setTimestamp(output.getTimestamp())
            self.out.send(detections_msg)



import depthai as dai

from ..messages.creators import create_depth_message


class MonocularDepthParser(dai.node.ThreadedHostNode):
    """Parser class for monocular depth models (e.g. Depth Anything model).

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    depth_type : str
        Type of depth output (relative or metric).
    depth_limit : float
        The maximum depth value (in meters) to be used in the depth map.

    Output Message/s
    ----------------
    **Type**: dai.ImgFrame

    **Description**: Depth message containing the depth map. The depth map is represented with dai.ImgFrame.

    Error Handling
    --------------
    **ValueError**: If the number of output layers is not E{1}.
    """

    def __init__(self, depth_type="relative", depth_limit=0.0):
        """Initializes the MonocularDepthParser node.

        @param depth_type: Type of depth output (relative or metric).
        @type depth_type: Literal['relative', 'metric']
        @param depth_limit: The maximum depth value (in meters) to be used in the depth
            map. The default value is 0, which means no limit.
        @type depth_limit: float
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.depth_type = depth_type
        self.depth_limit = depth_limit

    def setRelativeDepthType(self):
        """Sets the depth type to relative."""
        self.depth_type = "relative"

    def setMetricDepthType(self):
        """Sets the depth type to metric."""
        self.depth_type = "metric"

    def setDepthLimit(self, depth_limit):
        """Sets the depth limit."""
        self.depth_limit = depth_limit

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            output_layer_names = output.getAllLayerNames()
            if len(output_layer_names) != 1:
                raise ValueError(
                    f"Expected 1 output layer, got {len(output_layer_names)}."
                )

            output_map = output.getTensor(output_layer_names[0], dequantize=True)

            if len(output_map.shape) == 3:
                if output_map.shape[0] == 1:
                    depth_map = output_map[0]
                elif output_map.shape[2] == 1:
                    depth_map = output_map[:, :, 0]
            elif len(output_map.shape) == 2:
                depth_map = output_map
            elif len(output_map.shape) == 4:
                depth_map = output_map[0][0]
            else:
                raise ValueError(
                    f"Expected 3- or 2-dimensional output, got {len(output_map.shape)}-dimensional",
                )

            depth_message = create_depth_message(
                depth_map=depth_map,
                depth_type=self.depth_type,
                depth_limit=self.depth_limit,
            )
            depth_message.setTimestamp(output.getTimestamp())
            self.out.send(depth_message)



from typing import Optional, Tuple

import depthai as dai
import numpy as np

from ..messages.creators import create_sam_message
from .utils.fastsam import (
    box_prompt,
    decode_fastsam_output,
    point_prompt,
    process_single_mask,
)
from .yolo import YOLOExtendedParser


class FastSAMParser(YOLOExtendedParser):
    """Parser class for parsing the output of the FastSAM model.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    conf_threshold : float
        Confidence score threshold for detected faces.
    n_classes : int
        Number of classes in the model.
    iou_threshold : float
        Non-maximum suppression threshold.
    mask_conf : float
        Mask confidence threshold.
    input_shape : Tuple[int, int]
        Shape of the input image.
    prompt : str
        Prompt type.
    points : Tuple[int, int]
        Points.
    point_label : int
        Point label.
    bbox : Tuple[int, int, int, int]
        Bounding box.

    Output Message/s
    ----------------
    **Type**: SegmentationMasks

    **Description**: SegmentationMasks message containing the resulting segmentation masks given the prompt.

    Error Handling
    --------------
    """

    def __init__(
        self,
        conf_threshold: int = 0.5,
        n_classes: int = 1,
        iou_threshold: int = 0.5,
        mask_conf: float = 0.5,
        input_shape: Tuple[int, int] = (640, 640),
        prompt: str = "everything",
        points: Optional[Tuple[int, int]] = None,
        point_label: Optional[int] = None,
        bbox: Optional[Tuple[int, int, int, int]] = None,
    ):
        """Initialize the FastSAMParser node.

        @param conf_threshold: The confidence threshold for the detections
        @type conf_threshold: float
        @param n_classes: The number of classes in the model
        @type n_classes: int
        @param iou_threshold: The intersection over union threshold
        @type iou_threshold: float
        @param mask_conf: The mask confidence threshold
        @type mask_conf: float
        @param input_shape: The shape of the input image
        @type input_shape: Tuple[int, int]
        @param prompt: The prompt type
        @type prompt: str
        @param points: The points
        @type points: Optional[Tuple[int, int]]
        @param point_label: The point label
        @type point_label: Optional[int]
        @param bbox: The bounding box
        @type bbox: Optional[Tuple[int, int, int, int]]
        """
        YOLOExtendedParser.__init__(
            self, conf_threshold, n_classes, iou_threshold, mask_conf
        )
        self.input_shape = input_shape
        self.prompt = prompt
        self.points = points
        self.point_label = point_label
        self.bbox = bbox

    def setInputImageSize(self, width, height):
        """Sets the input image size.

        @param width: The width of the input image
        @type width: int
        @param height: The height of the input image
        @type height: int
        """
        self.input_shape = (width, height)

    def setPrompt(self, prompt):
        """Sets the prompt type.

        @param prompt: The prompt type
        @type prompt: str
        """
        self.prompt = prompt

    def setPoints(self, points):
        """Sets the points.

        @param points: The points
        @type points: Tuple[int, int]
        """
        self.points = points

    def setPointLabel(self, point_label):
        """Sets the point label.

        @param point_label: The point label
        @type point_label: int
        """
        self.point_label = point_label

    def setBoundingBox(self, bbox):
        """Sets the bounding box.

        @param bbox: The bounding box
        @type bbox: Tuple[int, int, int, int]
        """
        self.bbox = bbox

    def run(self):
        if self.prompt not in ["everything", "bbox", "point"]:
            raise ValueError("Prompt must be one of 'everything', 'bbox', or 'point'")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data
            # Get all the layer names
            layer_names = output.getAllLayerNames()

            outputs_names = sorted([name for name in layer_names if "_yolo" in name])
            outputs_values = [
                output.getTensor(o, dequantize=True).astype(np.float32)
                for o in outputs_names
            ]
            # Get the segmentation outputs
            (
                masks_outputs_values,
                protos_output,
                protos_len,
            ) = self._get_segmentation_outputs(output)

            if len(outputs_values[0].shape) != 4:
                # RVC4
                outputs_values = [
                    o.transpose((2, 0, 1))[np.newaxis, ...] for o in outputs_values
                ]
                (
                    protos_output,
                    protos_len,
                    masks_outputs_values,
                ) = self._reshape_seg_outputs(
                    protos_output, protos_len, masks_outputs_values
                )

            # Decode the outputs
            results = decode_fastsam_output(
                outputs_values,
                [8, 16, 32],
                [None, None, None],
                img_shape=self.input_shape[::-1],
                conf_thres=self.conf_threshold,
                iou_thres=self.iou_threshold,
                num_classes=self.n_classes,
            )

            bboxes, masks = [], []
            for i in range(results.shape[0]):
                bbox, conf, label, seg_coeff = (
                    results[i, :4].astype(int),
                    results[i, 4],
                    results[i, 5].astype(int),
                    results[i, 6:].astype(int),
                )
                bboxes.append(bbox.tolist() + [conf, int(label)])
                hi, ai, xi, yi = seg_coeff
                mask_coeff = masks_outputs_values[hi][
                    0, ai * protos_len : (ai + 1) * protos_len, yi, xi
                ]
                mask = process_single_mask(
                    protos_output[0], mask_coeff, self.mask_conf, self.input_shape, bbox
                )
                masks.append(mask)

            results_bboxes = np.array(bboxes)
            results_masks = np.array(masks)

            if self.prompt == "bbox":
                results_masks = box_prompt(
                    results_masks, bbox=self.bbox, orig_shape=self.input_shape[::-1]
                )
            elif self.prompt == "point":
                results_masks = point_prompt(
                    results_bboxes,
                    results_masks,
                    points=self.points,
                    pointlabel=self.point_label,
                    orig_shape=self.input_shape[::-1],
                )

            segmentation_message = create_sam_message(results_masks)
            segmentation_message.setTimestamp(output.getTimestamp())

            self.out.send(segmentation_message)



from typing import List, Union

import depthai as dai
import numpy as np

from ...messages import HandKeypoints, Keypoints


def create_hand_keypoints_message(
    hand_keypoints: np.ndarray,
    handedness: float,
    confidence: float,
    confidence_threshold: float,
) -> HandKeypoints:
    """Create a DepthAI message for hand keypoints detection.

    @param hand_keypoints: Detected 3D hand keypoints of shape (N,3) meaning [...,[x, y, z],...].
    @type hand_keypoints: np.ndarray
    @param handedness: Handedness score of the detected hand (left: < 0.5, right > 0.5).
    @type handedness: float
    @param confidence: Confidence score of the detected hand.
    @type confidence: float
    @param confidence_threshold: Confidence threshold for the present hand.
    @type confidence_threshold: float

    @return: HandKeypoints message containing the detected hand keypoints, handedness, and confidence score.
    @rtype: HandKeypoints

    @raise ValueError: If the hand_keypoints are not a numpy array.
    @raise ValueError: If the hand_keypoints are not of shape (N,3).
    @raise ValueError: If the hand_keypoints 2nd dimension is not of size E{3}.
    @raise ValueError: If the handedness is not a float.
    @raise ValueError: If the confidence is not a float.
    @raise ValueError: If the confidence_threshold is not a float.
    """

    if not isinstance(hand_keypoints, np.ndarray):
        raise ValueError(
            f"hand_keypoints should be numpy array, got {type(hand_keypoints)}."
        )
    if len(hand_keypoints.shape) != 2:
        raise ValueError(
            f"hand_keypoints should be of shape (N,3) meaning [...,[x, y, z],...], got {hand_keypoints.shape}."
        )
    if hand_keypoints.shape[1] != 3:
        raise ValueError(
            f"hand_keypoints 2nd dimension should be of size 3 e.g. [x, y, z], got {hand_keypoints.shape[1]}."
        )
    if not isinstance(handedness, float):
        raise ValueError(f"Handedness should be float, got {type(handedness)}.")
    if not isinstance(confidence, float):
        raise ValueError(f"Confidence should be float, got {type(confidence)}.")
    if not isinstance(confidence_threshold, float):
        raise ValueError(
            f"Confidence threshold should be float, got {type(confidence_threshold)}."
        )

    if handedness < 0 or handedness > 1:
        raise ValueError(
            f"Handedness should be between 0 and 1, got handedness {handedness}."
        )

    if confidence < 0 or confidence > 1:
        raise ValueError(
            f"Confidence should be between 0 and 1, got confidence {confidence}."
        )
    if confidence_threshold < 0 or confidence_threshold > 1:
        raise ValueError(
            f"Confidence threshold should be between 0 and 1, got confidence threshold {confidence_threshold}."
        )

    hand_keypoints_msg = HandKeypoints()
    hand_keypoints_msg.handedness = handedness
    hand_keypoints_msg.confidence = confidence
    points = []
    if confidence >= confidence_threshold:
        for i in range(hand_keypoints.shape[0]):
            pt = dai.Point3f()
            pt.x = hand_keypoints[i][0]
            pt.y = hand_keypoints[i][1]
            pt.z = hand_keypoints[i][2]
            points.append(pt)
    hand_keypoints_msg.keypoints = points

    return hand_keypoints_msg


def create_keypoints_message(
    keypoints: Union[np.ndarray, List[List[float]]],
    scores: Union[np.ndarray, List[float]] = None,
    confidence_threshold: float = None,
) -> Keypoints:
    """Create a DepthAI message for the keypoints.

    @param keypoints: Detected 2D or 3D keypoints of shape (N,2 or 3) meaning [...,[x, y],...] or [...,[x, y, z],...].
    @type keypoints: np.ndarray or List[List[float]]
    @param scores: Confidence scores of the detected keypoints.
    @type scores: Optional[np.ndarray or List[float]]
    @param confidence_threshold: Confidence threshold of keypoint detections.
    @type confidence_threshold: Optional[float]

    @return: Keypoints message containing the detected keypoints.
    @rtype: Keypoints

    @raise ValueError: If the keypoints are not a numpy array or list.
    @raise ValueError: If the scores are not a numpy array or list.
    @raise ValueError: If scores and keypoints do not have the same length.
    @raise ValueError: If score values are not floats.
    @raise ValueError: If score values are not between 0 and 1.
    @raise ValueError: If the confidence threshold is not a float.
    @raise ValueError: If the confidence threshold is not between 0 and 1.
    @raise ValueError: If the keypoints are not of shape (N,2 or 3).
    @raise ValueError: If the keypoints 2nd dimension is not of size E{2} or E{3}.
    """

    if not isinstance(keypoints, (np.ndarray, list)):
        raise ValueError(
            f"Keypoints should be numpy array or list, got {type(keypoints)}."
        )

    if scores is not None:
        if not isinstance(scores, (np.ndarray, list)):
            raise ValueError(
                f"Scores should be numpy array or list, got {type(scores)}."
            )

        if len(keypoints) != len(scores):
            raise ValueError(
                "Keypoints and scores should have the same length. Got {} keypoints and {} scores.".format(
                    len(keypoints), len(scores)
                )
            )

        if not all(isinstance(score, (float, np.floating)) for score in scores):
            raise ValueError("Scores should only contain float values.")
        if not all(0 <= score <= 1 for score in scores):
            raise ValueError("Scores should only contain values between 0 and 1.")

    if confidence_threshold is not None:
        if not isinstance(confidence_threshold, float):
            raise ValueError(
                f"The confidence_threshold should be float, got {type(confidence_threshold)}."
            )

        if not (0 <= confidence_threshold <= 1):
            raise ValueError(
                f"The confidence_threshold should be between 0 and 1, got confidence_threshold {confidence_threshold}."
            )

    if len(keypoints) != 0:
        dimension = len(keypoints[0])
        if dimension != 2 and dimension != 3:
            raise ValueError(
                f"All keypoints should be of dimension 2 or 3, got dimension {dimension}."
            )

    if isinstance(keypoints, list):
        for keypoint in keypoints:
            if not isinstance(keypoint, list):
                raise ValueError(
                    f"Keypoints should be list of lists or np.array, got list of {type(keypoint)}."
                )
            if len(keypoint) != dimension:
                raise ValueError(
                    "All keypoints have to be of same dimension e.g. [x, y] or [x, y, z], got mixed inner dimensions."
                )
            for coord in keypoint:
                if not isinstance(coord, (float, np.floating)):
                    raise ValueError(
                        f"Keypoints inner list should contain only float, got {type(coord)}."
                    )

    keypoints = np.array(keypoints)
    if scores is not None:
        scores = np.array(scores)

    if len(keypoints) != 0:
        if len(keypoints.shape) != 2:
            raise ValueError(
                f"Keypoints should be of shape (N,2 or 3) got {keypoints.shape}."
            )

        use_3d = keypoints.shape[1] == 3

    keypoints_msg = Keypoints()
    points = []
    for i, keypoint in enumerate(keypoints):
        if scores is not None:
            if scores[i] < confidence_threshold:
                continue
        pt = dai.Point3f()
        pt.x = keypoint[0]
        pt.y = keypoint[1]
        pt.z = keypoint[2] if use_3d else 0
        points.append(pt)

    keypoints_msg.keypoints = points
    return keypoints_msg



from .decode_detections import decode_detections
from .denormalize import unnormalize_image
from .medipipe import generate_anchors_and_decode

__all__ = [
    "unnormalize_image",
    "decode_detections",
    "generate_anchors_and_decode",
]



import depthai as dai

from depthai_nodes.ml.messages import (
    AgeGender,
    Classifications,
    ImgDetectionsExtended,
    Keypoints,
    Lines,
)


def parse_detection_message(message: dai.ImgDetections):
    """Parses the detection message and returns the detections."""
    detections = message.detections
    return detections


def parse_line_detection_message(message: Lines):
    """Parses the line detection message and returns the lines."""
    lines = message.lines
    return lines


def parse_segmentation_message(message: dai.ImgFrame):
    """Parses the segmentation message and returns the mask."""
    mask = message.getFrame()
    mask = mask.reshape(mask.shape[0], mask.shape[1])
    return mask


def parse_keypoints_message(message: Keypoints):
    """Parses the keypoints message and returns the keypoints."""
    keypoints = message.keypoints
    return keypoints


def parse_classification_message(message: Classifications):
    """Parses the classification message and returns the classification."""
    classes = message.classes
    scores = message.scores
    return classes, scores


def parse_image_message(message: dai.ImgFrame):
    """Parses the image message and returns the image."""
    image = message.getFrame()
    return image


def parser_age_gender_message(message: AgeGender):
    """Parses the age-gender message and return the age and scores for all genders."""

    age = message.age
    gender = message.gender
    gender_scores = gender.scores
    gender_classes = gender.classes

    return age, gender_classes, gender_scores


def parse_yolo_kpts_message(message: ImgDetectionsExtended):
    """Parses the yolo keypoints message and returns the keypoints."""
    detections = message.detections
    return detections



import depthai as dai
import numpy as np

from ..messages.creators import create_detection_message
from .utils.yolo import decode_yolo_output, parse_kpts, process_single_mask


class YOLOExtendedParser(dai.node.ThreadedHostNode):
    """Parser class for parsing the output of the YOLO Instance Segmentation and Pose
    Estimation models.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    conf_threshold : float
        Confidence score threshold for detected faces.
    n_classes : int
        Number of classes in the model.
    iou_threshold : float
        Intersection over union threshold.
    mask_conf : float
        Mask confidence threshold.
    n_keypoints : int
        Number of keypoints in the model.

    Output Message/s
    ----------------
    **Type**: ImgDetectionsExtended

    **Description**: Message containing bounding boxes, labels, confidence scores, and keypoints or masks and protos of the detected objects.
    """

    _KPTS_MODE = 0
    _SEG_MODE = 1

    def __init__(
        self,
        conf_threshold: int = 0.5,
        n_classes: int = 1,
        iou_threshold: int = 0.5,
        mask_conf: float = 0.5,
        n_keypoints: int = 17,
    ):
        """Initialize the YOLOExtendedParser node.

        @param conf_threshold: The confidence threshold for the detections
        @type conf_threshold: float
        @param n_classes: The number of classes in the model
        @type n_classes: int
        @param iou_threshold: The intersection over union threshold
        @type iou_threshold: float
        @param mask_conf: The mask confidence threshold
        @type mask_conf: float
        @param n_keypoints: The number of keypoints in the model
        @type n_keypoints: int
        """
        dai.node.ThreadedHostNode.__init__(self)
        self.input = self.createInput()
        self.out = self.createOutput()

        self.conf_threshold = conf_threshold
        self.n_classes = n_classes
        self.iou_threshold = iou_threshold
        self.mask_conf = mask_conf
        self.n_keypoints = n_keypoints

    def setConfidenceThreshold(self, threshold):
        """Sets the confidence score threshold for detected faces.

        @param threshold: Confidence score threshold for detected faces.
        @type threshold: float
        """
        self.conf_threshold = threshold

    def setNumClasses(self, n_classes):
        """Sets the number of classes in the model.

        @param numClasses: The number of classes in the model.
        @type numClasses: int
        """
        self.n_classes = n_classes

    def setIouThreshold(self, iou_threshold):
        """Sets the intersection over union threshold.

        @param iou_threshold: The intersection over union threshold.
        @type iou_threshold: float
        """
        self.iou_threshold = iou_threshold

    def setMaskConfidence(self, mask_conf):
        """Sets the mask confidence threshold.

        @param mask_conf: The mask confidence threshold.
        @type mask_conf: float
        """
        self.mask_conf = mask_conf

    def setNumKeypoints(self, n_keypoints):
        """Sets the number of keypoints in the model.

        @param n_keypoints: The number of keypoints in the model.
        @type n_keypoints: int
        """
        self.n_keypoints = n_keypoints

    def _get_segmentation_outputs(self, output):
        """Get the segmentation outputs from the Neural Network data."""
        # Get all the layer names
        layer_names = output.getAllLayerNames()
        mask_outputs = sorted([name for name in layer_names if "_masks" in name])
        masks_outputs_values = [
            output.getTensor(o, dequantize=True).astype(np.float32)
            for o in mask_outputs
        ]
        protos_output = output.getTensor("protos_output", dequantize=True).astype(
            np.float32
        )
        protos_len = protos_output.shape[1]
        return masks_outputs_values, protos_output, protos_len

    def _reshape_seg_outputs(self, protos_output, protos_len, masks_outputs_values):
        """Reshape the segmentation outputs."""
        protos_output = protos_output.transpose((2, 0, 1))[np.newaxis, ...]
        protos_len = protos_output.shape[1]
        masks_outputs_values = [
            o.transpose((2, 0, 1))[np.newaxis, ...] for o in masks_outputs_values
        ]
        return protos_output, protos_len, masks_outputs_values

    def run(self):
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data
            # Get all the layer names
            layer_names = output.getAllLayerNames()

            outputs_names = sorted([name for name in layer_names if "_yolo" in name])
            outputs_values = [
                output.getTensor(o, dequantize=True).astype(np.float32)
                for o in outputs_names
            ]

            if any("kpt_output" in name for name in layer_names):
                mode = self._KPTS_MODE
                # Get the keypoint outputs
                kpts_output_names = sorted(
                    [name for name in layer_names if "kpt_output" in name]
                )
                kpts_outputs = [
                    output.getTensor(o, dequantize=True).astype(np.float32)
                    for o in kpts_output_names
                ]
            elif any("_masks" in name for name in layer_names):
                mode = self._SEG_MODE
                # Get the segmentation outputs
                (
                    masks_outputs_values,
                    protos_output,
                    protos_len,
                ) = self._get_segmentation_outputs(output)

            if len(outputs_values[0].shape) != 4:
                # RVC4
                outputs_values = [
                    o.transpose((2, 0, 1))[np.newaxis, ...] for o in outputs_values
                ]
                if mode == self._KPTS_MODE:
                    kpts_outputs = [o[np.newaxis, ...] for o in kpts_outputs]
                elif mode == self._SEG_MODE:
                    (
                        protos_output,
                        protos_len,
                        masks_outputs_values,
                    ) = self._reshape_seg_outputs(
                        protos_output, protos_len, masks_outputs_values
                    )

            # Decode the outputs
            results = decode_yolo_output(
                outputs_values,
                [8, 16, 32],
                [None, None, None],
                kpts=kpts_outputs if mode == self._KPTS_MODE else None,
                conf_thres=self.conf_threshold,
                iou_thres=self.iou_threshold,
                num_classes=self.n_classes,
            )

            bboxes, labels, scores, additional_output = [], [], [], []
            for i in range(results.shape[0]):
                bbox, conf, label, other = (
                    results[i, :4].astype(int),
                    results[i, 4],
                    results[i, 5].astype(int),
                    results[i, 6:],
                )

                bboxes.append(bbox)
                labels.append(int(label))
                scores.append(conf)

                if mode == self._KPTS_MODE:
                    kpts = parse_kpts(other, self.n_keypoints)
                    additional_output.append(kpts)
                elif mode == self._SEG_MODE:
                    seg_coeff = other.astype(int)
                    hi, ai, xi, yi = seg_coeff
                    mask_coeff = masks_outputs_values[hi][
                        0, ai * protos_len : (ai + 1) * protos_len, yi, xi
                    ]
                    mask = process_single_mask(
                        protos_output[0], mask_coeff, self.mask_conf
                    )
                    additional_output.append(mask)

            if mode == self._KPTS_MODE:
                detections_message = create_detection_message(
                    np.array(bboxes),
                    np.array(scores),
                    labels,
                    keypoints=additional_output,
                )
            elif mode == self._SEG_MODE:
                detections_message = create_detection_message(
                    np.array(bboxes), np.array(scores), labels, masks=additional_output
                )
            detections_message.setTimestamp(output.getTimestamp())

            self.out.send(detections_message)



import depthai as dai
from utils.arguments import initialize_argparser, parse_model_slug
from utils.model import get_input_shape, get_model_from_hub, get_parser
from utils.parser import setup_parser
from visualization.visualize import visualize

# Initialize the argument parser
arg_parser, args = initialize_argparser()

# Parse the model slug
model_slug, model_version_slug = parse_model_slug(args)

# Get the model from the HubAI
nn_archive = get_model_from_hub(model_slug, model_version_slug)

# Get the parser
parser_class, parser_name = get_parser(nn_archive)
input_shape = get_input_shape(nn_archive)

if parser_name == "XFeatParser":
    raise NotImplementedError("XFeatParser is not supported in this script yet.")

# Create the pipeline
with dai.Pipeline() as pipeline:
    cam = pipeline.create(dai.node.Camera).build()

    # YOLO and MobileNet-SSD have native parsers in DAI - no need to create a separate parser
    if parser_name == "YOLO" or parser_name == "SSD":
        network = pipeline.create(dai.node.DetectionNetwork).build(
            cam.requestOutput(input_shape, type=dai.ImgFrame.Type.BGR888p), nn_archive
        )
        parser_queue = network.out.createOutputQueue()
    else:
        image_type = dai.ImgFrame.Type.BGR888p
        if "gray" in model_version_slug:
            image_type = dai.ImgFrame.Type.GRAY8

        if input_shape[0] < 128 or input_shape[1] < 128:
            print(
                "Input shape is too small so we are requesting a larger image and resizing it."
            )
            print(
                "During visualization we resize small image back to large, so image quality is lower."
            )
            manip = pipeline.create(dai.node.ImageManip)
            manip.initialConfig.setResize(input_shape)
            large_input_shape = (input_shape[0] * 4, input_shape[1] * 4)
            cam.requestOutput(large_input_shape, type=image_type).link(manip.inputImage)
            network = pipeline.create(dai.node.NeuralNetwork).build(
                manip.out, nn_archive
            )
        else:
            network = pipeline.create(dai.node.NeuralNetwork).build(
                cam.requestOutput(input_shape, type=image_type), nn_archive
            )

        parser = pipeline.create(parser_class)
        setup_parser(parser, nn_archive, parser_name)

        # Linking
        network.out.link(parser.input)

        parser_queue = parser.out.createOutputQueue()

    camera_queue = network.passthrough.createOutputQueue()

    pipeline.start()

    while pipeline.isRunning():
        frame: dai.ImgFrame = camera_queue.get().getCvFrame()
        message = parser_queue.get()

        extraParams = (
            nn_archive.getConfig().getConfigV1().model.heads[0].metadata.extraParams
        )
        if visualize(frame, message, parser_name, extraParams):
            pipeline.stop()
            break



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.ml.messages import HandKeypoints
from depthai_nodes.ml.messages.creators.keypoints import create_hand_keypoints_message


def test_none_hand_keypoints():
    with pytest.raises(ValueError):
        create_hand_keypoints_message(None, 0.5, 0.8, 0.7)


def test_not_nparray_hand_keypoints():
    with pytest.raises(ValueError):
        create_hand_keypoints_message([[1, 2, 3]], 0.5, 0.8, 0.7)


def test_empty_nparray():
    with pytest.raises(ValueError):
        create_hand_keypoints_message(np.array([]), 0.5, 0.8, 0.7)


def test_3d_empty_nparray():
    with pytest.raises(ValueError):
        create_hand_keypoints_message(np.array([[[]]]), 0.5, 0.8, 0.7)


def test_wrong_dimension_nparray():
    with pytest.raises(ValueError):
        create_hand_keypoints_message(np.array([[], [], []]), 0.5, 0.8, 0.7)


def test_handedness_float():
    with pytest.raises(
        ValueError, match="Handedness should be float, got <class 'str'>."
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), "0.5", 0.8, 0.7)


def test_confidence_float():
    with pytest.raises(
        ValueError, match="Confidence should be float, got <class 'str'>."
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, "0.8", 0.7)


def test_confidence_threshold_float():
    with pytest.raises(
        ValueError, match="Confidence threshold should be float, got <class 'str'>."
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 0.8, "0.7")


def test_low_confidence():
    create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 1e-10, 1e-11)


def test_negative_confidence():
    with pytest.raises(
        ValueError, match="Confidence should be between 0 and 1, got confidence -0.8."
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, -0.8, 0.7)


def test_negative_confidence_threshold():
    with pytest.raises(
        ValueError,
        match="Confidence threshold should be between 0 and 1, got confidence threshold -0.7.",
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 0.8, -0.7)


def test_big_confidence():
    with pytest.raises(
        ValueError, match="Confidence should be between 0 and 1, got confidence 500.0."
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 500.0, 0.7)


def test_big_confidence_threshold():
    with pytest.raises(
        ValueError,
        match="Confidence threshold should be between 0 and 1, got confidence threshold 500.0.",
    ):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 0.8, 500.0)


def test_handedness_between_0_and_1():
    with pytest.raises(ValueError):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), 1.5, 0.8, 0.7)
    with pytest.raises(ValueError):
        create_hand_keypoints_message(np.array([[1, 2, 3]]), -0.5, 0.8, 0.7)


def test_confidence_below_thr():
    message = create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 0.8, 0.9)

    assert isinstance(message, HandKeypoints)
    assert message.confidence == 0.8
    assert message.handedness == 0.5
    assert message.keypoints == []


def test_confidence_above_thr():
    message = create_hand_keypoints_message(np.array([[1, 2, 3]]), 0.5, 0.8, 0.7)

    true_pt = dai.Point3f(1, 2, 3)
    true_pt.x = 1
    true_pt.y = 2
    true_pt.z = 3

    assert message.confidence == 0.8
    assert message.handedness == 0.5
    assert isinstance(message.keypoints, list)
    assert isinstance(message.keypoints[0], dai.Point3f)
    assert message.keypoints[0].x == true_pt.x
    assert message.keypoints[0].y == true_pt.y
    assert message.keypoints[0].z == true_pt.z


def test_list_keypoints():
    message = create_hand_keypoints_message(
        np.array([[1, 2, 3], [4, 5, 6]]), 0.5, 0.8, 0.7
    )

    assert len(message.keypoints) == 2
    assert message.keypoints[0].x == 1
    assert message.keypoints[0].y == 2
    assert message.keypoints[0].z == 3
    assert message.keypoints[1].x == 4
    assert message.keypoints[1].y == 5
    assert message.keypoints[1].z == 6


if __name__ == "__main__":
    pytest.main()



from .classification import visualize_age_gender, visualize_classification
from .detection import (
    visualize_detections,
    visualize_line_detections,
    visualize_yolo_extended,
)
from .image import visualize_image
from .keypoints import visualize_keypoints
from .segmentation import visualize_segmentation

parser_mapping = {
    "YuNetParser": visualize_detections,
    "SCRFDParser": visualize_detections,
    "MPPalmDetectionParser": visualize_detections,
    "YOLO": visualize_detections,
    "SSD": visualize_detections,
    "SegmentationParser": visualize_segmentation,
    "MLSDParser": visualize_line_detections,
    "KeypointParser": visualize_keypoints,
    "HRNetParser": visualize_keypoints,
    "SuperAnimalParser": visualize_keypoints,
    "MPHandLandmarkParser": visualize_keypoints,
    "ClassificationParser": visualize_classification,
    "ImageOutputParser": visualize_image,
    "MonocularDepthParser": visualize_image,
    "AgeGenderParser": visualize_age_gender,
    "YOLOExtendedParser": visualize_yolo_extended,
}



from typing import List

import depthai as dai


class Line(dai.Buffer):
    """Line class for storing a line.

    Attributes
    ----------
    start_point : dai.Point2f
        Start point of the line with x and y coordinate.
    end_point : dai.Point2f
        End point of the line with x and y coordinate.
    confidence : float
        Confidence of the line.
    """

    def __init__(self):
        """Initializes the Line object."""
        super().__init__()
        self._start_point: dai.Point2f = None
        self._end_point: dai.Point2f = None
        self._confidence: float = None

    @property
    def start_point(self) -> dai.Point2f:
        """Returns the start point of the line.

        @return: Start point of the line.
        @rtype: dai.Point2f
        """
        return self._start_point

    @start_point.setter
    def start_point(self, value: dai.Point2f):
        """Sets the start point of the line.

        @param value: Start point of the line.
        @type value: dai.Point2f
        @raise TypeError: If the start point is not of type dai.Point2f.
        """
        if not isinstance(value, dai.Point2f):
            raise TypeError(
                f"start_point must be of type Point2f, instead got {type(value)}."
            )
        self._start_point = value

    @property
    def end_point(self) -> dai.Point2f:
        """Returns the end point of the line.

        @return: End point of the line.
        @rtype: dai.Point2f
        """
        return self._end_point

    @end_point.setter
    def end_point(self, value: dai.Point2f):
        """Sets the end point of the line.

        @param value: End point of the line.
        @type value: dai.Point2f
        @raise TypeError: If the end point is not of type dai.Point2f.
        """
        if not isinstance(value, dai.Point2f):
            raise TypeError(
                f"end_point must be of type Point2f, instead got {type(value)}."
            )
        self._end_point = value

    @property
    def confidence(self) -> float:
        """Returns the confidence of the line.

        @return: Confidence of the line.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the line.

        @param value: Confidence of the line.
        @type value: float
        @raise TypeError: If the confidence is not of type float.
        """
        if not isinstance(value, float):
            raise TypeError(
                f"confidence must be of type float, instead got {type(value)}."
            )
        self._confidence = value


class Lines(dai.Buffer):
    """Lines class for storing lines.

    Attributes
    ----------
    lines : List[Line]
        List of detected lines.
    """

    def __init__(self):
        """Initializes the Lines object."""
        super().__init__()
        self._lines: List[Line] = []

    @property
    def lines(self) -> List[Line]:
        """Returns the lines.

        @return: List of lines.
        @rtype: List[Line]
        """
        return self._lines

    @lines.setter
    def lines(self, value: List[Line]):
        """Sets the lines.

        @param value: List of lines.
        @type value: List[Line]
        @raise TypeError: If the lines are not a list.
        @raise TypeError: If each line is not of type Line.
        """
        if not isinstance(value, List):
            raise TypeError(
                f"lines must be of type List[Line], instead got {type(value)}."
            )
        for line in value:
            if not isinstance(line, Line):
                raise TypeError(
                    f"lines must be of type List[Line], instead got {type(value)}."
                )
        self._lines = value



from typing import Literal

import depthai as dai
import numpy as np

UINT16_MAX_VALUE = 65535


def create_depth_message(
    depth_map: np.ndarray,
    depth_type: Literal["relative", "metric"],
    depth_limit: float = 0.0,
) -> dai.ImgFrame:
    """Create a DepthAI message for a depth map.

    @param depth_map: A NumPy array representing the depth map with shape HW or NHW/HWN.
        Here N stands for batch dimension.
    @type depth_map: np.array
    @param depth_type: A string indicating the type of depth map. It can either be
        'relative' or 'metric'.
    @type depth_type: Literal['relative', 'metric']
    @param depth_limit: The maximum depth value (in meters) to be used in the depth map.
        The default value is 0, which means no limit.
    @type depth_limit: float
    @return: An ImgFrame object containing the depth information.
    @rtype: dai.ImgFrame
    @raise ValueError: If the depth map is not a NumPy array.
    @raise ValueError: If the depth map is not 2D or 3D.
    @raise ValueError: If the depth map shape is not NHW or HWN.
    @raise ValueError: If the depth type is not 'relative' or 'metric'.
    @raise ValueError: If the depth limit is not 0 and the depth type is 'relative'.
    @raise ValueError: If the depth limit is 0 and the depth type is 'metric'.
    @raise ValueError: If the depth limit is negative.
    """

    if not isinstance(depth_map, np.ndarray):
        raise ValueError(f"Expected numpy array, got {type(depth_map)}.")

    if len(depth_map.shape) == 3:
        if depth_map.shape[0] == 1:
            depth_map = depth_map[0, :, :]  # NHW to HW
        elif depth_map.shape[2] == 1:
            depth_map = depth_map[:, :, 0]  # HWN to HW
        else:
            raise ValueError(
                f"Unexpected image shape. Expected NHW or HWN, got {depth_map.shape}."
            )

    if len(depth_map.shape) != 2:
        raise ValueError(f"Expected 2D or 3D input, got {len(depth_map.shape)}D input.")

    if not (depth_type == "relative" or depth_type == "metric"):
        raise ValueError(
            f"Invalid depth type: {depth_type}. Only 'relative' and 'metric' are supported."
        )

    if depth_type == "relative" and depth_limit != 0:
        raise ValueError(
            f"Invalid depth limit: {depth_limit}. For relative depth, depth limit must be equal to 0."
        )

    if depth_type == "metric" and depth_limit == 0:
        raise ValueError(
            f"Invalid depth limit: {depth_limit}. For metric depth, depth limit must be bigger than 0."
        )

    if depth_limit < 0:
        raise ValueError(
            f"Invalid depth limit: {depth_limit}. Depth limit must be bigger than 0."
        )

    data_type = dai.ImgFrame.Type.RAW16

    min_val = depth_map.min() if depth_type == "relative" else 0
    max_val = depth_map.max() if depth_type == "relative" else depth_limit

    # clip values bigger than max_val
    depth_map = np.clip(depth_map, a_min=None, a_max=max_val)

    # normalize depth map to UINT16 range [0, UINT16_MAX_VALUE]
    if min_val == max_val:  # avoid division by zero
        depth_map = np.zeros_like(depth_map)
    else:
        depth_map = (depth_map - min_val) / (max_val - min_val) * UINT16_MAX_VALUE
    depth_map = depth_map.astype(np.uint16)

    imgFrame = dai.ImgFrame()
    imgFrame.setFrame(depth_map)
    imgFrame.setWidth(depth_map.shape[1])
    imgFrame.setHeight(depth_map.shape[0])
    imgFrame.setType(data_type)

    return imgFrame
